from __future__ import annotations

import argparse
import importlib.metadata
import logging
import os
import site
import subprocess
import sys
import warnings
from collections.abc import Mapping
from itertools import product
from pathlib import Path
from typing import Any, Literal, TypeAlias

from ConfigSpace import ConfigurationSpace
from hpoglue import BenchmarkDescription, Config, FunctionalBenchmark, Optimizer, Problem
from hpoglue.utils import dict_to_configpriors
from packaging import version

OptWithHps: TypeAlias = tuple[type[Optimizer], Mapping[str, Any]]

BenchWith_Objs_Fids: TypeAlias = tuple[BenchmarkDescription, Mapping[str, Any]]

GLOBAL_SEED = 42

logger = logging.getLogger(__name__)


HPOSUITR_REPO = "github.com/automl/hposuite.git"
HPOSUITE_EDITABLE = Path(__file__).parent.parent
HPOSUITE_PYPI = "hposuite"
HPOSUITE_GIT_SSH_INSTALL = "git+ssh://git@github.com/automl/hposuite.git"

class GlueWrapperFunctions:
    """A collection of wrapper functions around certain hpoglue methods."""

    @staticmethod
    def problem_from_dict(data: dict[str, Any]) -> Problem:
        """Convert a dictionary to a Problem instance."""
        from hposuite.benchmarks import BENCHMARKS
        from hposuite.optimizers import OPTIMIZERS

        return Problem.from_dict(
            data=data,
            benchmarks_dict=BENCHMARKS,
            optimizers_dict=OPTIMIZERS,
        )


class HiddenPrints:  # noqa: D101
    def __enter__(self):
        self._original_stdout = sys.stdout
        from pathlib import Path
        sys.stdout = Path(os.devnull).open("w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout



def is_package_installed(package_name: str) -> bool:
    """Check if a package is installed in the current virtual environment."""
    # Extract version constraint if present
    version_constraints = ["==", ">=", "<=", ">", "<"]
    version_check = None
    for constraint in version_constraints:
        if constraint in package_name:
            package_name, version_spec = package_name.split(constraint)
            version_check = (constraint, version_spec)
            break

    # Normalize package name (replace hyphens with underscores)
    package_name = package_name.replace("-", "_")

    # Remove dependencies from the package name
    package_name = package_name.split("[")[0]

    # Get the site-packages directory of the current virtual environment
    venv_site_packages = site.getsitepackages() if hasattr(site, "getsitepackages") else []
    venv_prefix = sys.prefix  # Virtual environment root

    # Check if the package is installed in the virtual environment
    for site_package_path in venv_site_packages:
        package_path = Path(site_package_path) / package_name

        # Check if the package exists in the site-packages directory
        if package_path.exists() and venv_prefix in str(package_path):
            installed_version = importlib.metadata.version(package_name)
            if version_check:
                constraint, required_version = version_check
                return _check_package_version(installed_version, required_version, constraint)
            return True

        # Check if package is installed as different name (e.g., .dist-info or .egg-info)
        dist_info_pattern = f"{package_name}*"
        dist_info_paths = list(Path(site_package_path).glob(dist_info_pattern))
        if dist_info_paths:
            dist_info_name = dist_info_paths[0].name.replace(".dist-info", "") \
                .replace(".egg-info", "")
            installed_version = dist_info_name.split("-")[-1]
            if version_check:
                constraint, required_version = version_check
                return _check_package_version(installed_version, required_version, constraint)
            return True

    return False

def _check_package_version(
    installed_version: str,
    required_version: str,
    check_key: str,
):
    """Check if the installed package version satisfies the required version."""
    installed_version = version.parse(installed_version)
    required_version = version.parse(required_version)
    match check_key:
        case "==":
            return installed_version == required_version
        case ">=":
            return installed_version >= required_version
        case "<=":
            return installed_version <= required_version
        case ">":
            return installed_version > required_version
        case "<":
            return installed_version < required_version


def get_current_installed_hposuite_version() -> str:
    """Retrieve the currently installed version of hposuite."""
    cmd = ["pip", "show", "hposuite"]
    logger.debug(cmd)
    output = subprocess.run(  # noqa: S603
        cmd,
        check=True,
        capture_output=True,
        text=True,
    )
    lines = output.stdout.strip().splitlines()
    for line in lines:
        if "Version: " in line:
            return line.split(": ")[1]

    raise RuntimeError(f"Could not find hposuite version in {lines}.")


def get_current_ConfigSpace_version() -> version.Version:  # noqa: N802
    """Retrieve the currently installed version of ConfigSpace."""
    pkg_dict = {dist.name: dist.version for dist in importlib.metadata.distributions()}
    if "ConfigSpace" in pkg_dict:
        return version.parse(pkg_dict["ConfigSpace"])
    raise RuntimeError("ConfigSpace is not installed.")


def compare_installed_CS_version_vs_required(required_version: str) -> str:  # noqa: N802
    """Compare the installed ConfigSpace version with the required version."""
    installed_version = get_current_ConfigSpace_version()
    required_version = version.parse(required_version)
    if installed_version < required_version:
        return "<"
    if installed_version == required_version:
        return "=="
    return ">"


def get_compatible(  # noqa: C901, PLR0912, PLR0915
    *,
    optimizers: (
        str
        | tuple[str, Mapping[str, Any]]
        | type[Optimizer]
        | OptWithHps # tuple[type[Optimizer], Mapping[str, Any]]
        | list[tuple[str, Mapping[str, Any]]]
        | list[str]
        | list[OptWithHps] # list[tuple[type[Optimizer], Mapping[str, Any]]]
        | list[type[Optimizer]]
    ),
    benchmarks: (
        str
        | BenchmarkDescription
        | FunctionalBenchmark
        | tuple[str, Mapping[str, Any]]
        | BenchWith_Objs_Fids # tuple[BenchmarkDescription, Mapping[str, Any]]
        | tuple[FunctionalBenchmark, Mapping[str, Any]]
        | list[str]
        | list[BenchmarkDescription]
        | list[FunctionalBenchmark]
        | list[tuple[str, Mapping[str, Any]]]
        | list[BenchWith_Objs_Fids]   # list[tuple[BenchmarkDescription, Mapping[str, Any]]]
        | list[tuple[FunctionalBenchmark, Mapping[str, Any]]]
    ),
) -> None:
    """Get compatible optimizers and benchmark pairs from lists of optimizers and benchmarks.

    Args:
        optimizers: List of optimizers to check.
                    "all" to check all optimizers.

        benchmarks: List of benchmarks to check.
                    "all" to check all benchmarks.
    """
    from hposuite.benchmarks import BENCHMARKS
    from hposuite.optimizers import OPTIMIZERS
    from hposuite.run import Run

    assert optimizers, "At least one optimizer must be provided!"
    if not isinstance(optimizers, list):
        optimizers = [optimizers]

    assert benchmarks, "At least one benchmark must be provided!"
    if not isinstance(benchmarks, list):
        benchmarks = [benchmarks]


    _optimizers: list[OptWithHps] = []
    for optimizer in optimizers:
        match optimizer:
            case "all":
                _optimizers = [(opt, {}) for opt in OPTIMIZERS.values()]
                break
            case str():
                assert optimizer in OPTIMIZERS, (
                    f"Optimizer must be one of {OPTIMIZERS.keys()}\n"
                    f"Found {optimizer}"
                )
                _optimizers.append((OPTIMIZERS[optimizer], {}))
            case tuple():
                opt, hps = optimizer
                match opt:
                    case str():
                        assert opt in OPTIMIZERS, (
                            f"Optimizer must be one of {OPTIMIZERS.keys()}\n"
                            f"Found {opt}"
                        )
                        _optimizers.append((OPTIMIZERS[opt], hps))
                    case type():
                        _optimizers.append((opt, hps))
                    case _:
                        raise TypeError(f"Unknown Optimizer type {type(opt)}")
            case type():
                _optimizers.append(optimizer)
            case _:
                raise TypeError(f"Unknown Optimizer type {type(optimizers)}")


    _benchmarks: list[BenchWith_Objs_Fids] = []
    for benchmark in benchmarks:
        match benchmark:
            case "all":
                _benchmarks = [
                    (
                        bench if not isinstance(bench, FunctionalBenchmark) else bench.desc,
                        {}
                    )
                    for bench in BENCHMARKS.values()
                ]
                break
            case str():
                assert benchmark in BENCHMARKS, (
                    f"Benchmark must be one of {BENCHMARKS.keys()}\n"
                    f"Found {benchmark}"
                )
                if not isinstance(BENCHMARKS[benchmark], FunctionalBenchmark):
                    _benchmarks.append((BENCHMARKS[benchmark], {}))
                else:
                    _benchmarks.append((BENCHMARKS[benchmark].desc, {}))
            case BenchmarkDescription():
                _benchmarks.append((benchmark, {}))
            case FunctionalBenchmark():
                _benchmarks.append((benchmark.desc, {}))
            case tuple():
                bench, bench_hps = benchmark
                match bench:
                    case str():
                        assert bench in BENCHMARKS, (
                            f"Benchmark must be one of {BENCHMARKS.keys()}\n"
                            f"Found {bench}"
                        )
                        if not isinstance(BENCHMARKS[bench], FunctionalBenchmark):
                            _benchmarks.append((BENCHMARKS[bench], bench_hps))
                        else:
                            _benchmarks.append((BENCHMARKS[bench].desc, bench_hps))
                    case BenchmarkDescription():
                        _benchmarks.append((bench, bench_hps))
                    case FunctionalBenchmark():
                        _benchmarks.append((bench.desc, bench_hps))
                    case _:
                        raise TypeError(f"Unknown Benchmark type {type(benchmark)}")
            case _:
                raise TypeError(f"Unknown Benchmark type {type(benchmarks)}")


    _problems: list[Problem] = []
    for (opt, hps), (bench, objs_fids) in product(_optimizers, _benchmarks):
        try:

            objectives: int | str | list[str]
            fidelities: int | str | list[str] | None
            costs: int | str | list[str] | None

            objectives = objs_fids.get("objectives", 1)
            if isinstance(objectives, list) and len(objectives) == 1:
                objectives = objectives[0]

            fidelities = objs_fids.get("fidelities", None)
            if isinstance(fidelities, list) and len(fidelities) == 1:
                fidelities = fidelities[0]

            costs = objs_fids.get("costs", None)
            if isinstance(costs, list) and len(costs) == 1:
                costs = costs[0]

            priors = objs_fids.get("priors")

            _compatible_pairs = []
            _incompatible_pairs = []

            match fidelities, bench.fidelities:
                case None, None:
                    fidelities = None
                case None, _:
                    match opt.support.fidelities[0]:
                        case "single":
                            fidelities = 1
                        case "many":
                            fidelities = len(bench.fidelities)
                        case None:
                            fidelities = None
                        case _:
                            raise ValueError("Invalid fidelity support")
                case str() | int() | list(), _:
                    match opt.support.fidelities[0]:
                        case str():
                            pass
                        case None:
                            fidelities = None
                        case _:
                            raise ValueError("Invalid fidelity support")
                case _:
                    raise ValueError(
                        f"Invalid fidelity type: {type(fidelities)}. "
                        f"Expected None, str, int or list"
                    )

            if priors:
                priors = dict_to_configpriors(priors)


            _problem = Problem.problem(
                optimizer=opt,
                optimizer_hyperparameters=hps,
                benchmark=bench,
                objectives=objectives,
                budget=10,
                fidelities=fidelities,
                costs=costs,
                priors=priors,
            )
            _problems.append(_problem)
        except ValueError as e:
            warnings.warn(f"{e}\nTo ignore this, set `on_error='ignore'`", stacklevel=2)
            continue


    # NOTE: REDUNDANCY CHECK: _runs_per_problem is a dict now to avoid duplicate runs esp.
    # for eg. in case a BO Opt being used in combination with a
    # MF Opt on a benchmark with multiple fidelities -> explicitly adding different fidelities
    # in multiple benchmark instances would create redundant problems if BO Opts are present.
    _runs_per_problem: Mapping[str, Run] = {}
    for _problem in _problems:
        try:
            _run = Run(
                    problem=_problem,
                    seed=1,
                )
            if _run.name not in _runs_per_problem:
                _runs_per_problem[_run.name] = _run
                _compatible_pairs.append(
                    (
                        _problem.optimizer.name,
                        _problem.benchmark.name,
                    )
                )
        except ValueError as e:
            warnings.warn(f"{e}\nTo ignore this, set `on_error='ignore'`", stacklevel=2)
            continue


    match _compatible_pairs, len(_optimizers), len(_benchmarks):
        case [], _, _:
            logger.warning(
                "No compatible pairs found. "
            )
        case _, 1, _:
            opt = _optimizers[0][0].name
            benches = [
                bench for _, bench in _compatible_pairs
            ]
            print(  # noqa: T201
                f"Compatible Benchmarks for Optimizer `{opt}`:\n"
                f"{benches}"
            )
        case _, _, 1:
            bench = _benchmarks[0][0].name
            opts = [
                opt for opt, _ in _compatible_pairs
            ]
            print(  # noqa: T201
                f"Compatible Optimizers for Benchmark `{bench}`:\n"
                f"{opts}"
            )
        case _:
            print(  # noqa: T201
                "Compatible Optimizer-Benchmark pairs:\n"
                f"{_compatible_pairs}"
            )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Get compatible optimizers and benchmarks.")
    subparsers = parser.add_subparsers(dest="command")
    get_compatible_parser = subparsers.add_parser(
        "get_compatible",
        help="Get compatible optimizers and benchmarks.",
    )
    get_compatible_parser.add_argument(
        "--optimizers", "-opts",
        type=str,
        nargs="+",
        help="List of optimizers to check. Use 'all' to check all optimizers.",
    )
    get_compatible_parser.add_argument(
        "--benchmarks", "-benches",
        type=str,
        nargs="+",
        help="List of benchmarks to check. Use 'all' to check all benchmarks.",
    )
    args = parser.parse_args()

    get_compatible(optimizers=args.optimizers, benchmarks=args.benchmarks)


def set_priors_as_defaults(  # noqa: C901, PLR0912
    config_space: ConfigurationSpace,
    priors: Config,
    seed: int = 0,
    distribution: Literal["uniform", "normal", "beta"] = "uniform",
    sigma: float = 0.25,
    alpha: float = 2.0,
    beta: float = 2.0,
) -> ConfigurationSpace:
    """Set priors as defaults in the configuration space."""
    from ConfigSpace.hyperparameters import (
        BetaFloatHyperparameter,
        BetaIntegerHyperparameter,
        CategoricalHyperparameter,
        Constant,
        FloatHyperparameter,
        IntegerHyperparameter,
        NormalFloatHyperparameter,
        NormalIntegerHyperparameter,
        OrdinalHyperparameter,
        UniformFloatHyperparameter,
        UniformIntegerHyperparameter,
    )
    priors = priors.values
    assert len(priors) == len(list(config_space.values())), (
        f"Number of hyperparameters in prior ({len(priors)}) "
        f"must match number of hyperparameters in config space ({len(config_space.values())})."
    )
    cs = ConfigurationSpace(seed=seed)
    for hp in list(config_space.values()):
        assert hp.name in priors
        match hp:
            case FloatHyperparameter():
                match distribution:
                    case "uniform":
                        cs.add(
                            UniformFloatHyperparameter(
                                name=hp.name,
                                lower=hp.lower,
                                upper=hp.upper,
                                default_value=priors[hp.name],
                                log=hp.log
                            )
                        )
                    case "normal":
                        cs.add(
                            NormalFloatHyperparameter(
                                name=hp.name,
                                mu=priors[hp.name],
                                sigma=sigma,
                                lower=hp.lower,
                                upper=hp.upper,
                                default_value=priors[hp.name],
                                log=hp.log
                            )
                        )
                    case "beta":
                        cs.add(
                            BetaFloatHyperparameter(
                                name=hp.name,
                                alpha=alpha,
                                beta=beta,
                                lower=hp.lower,
                                upper=hp.upper,
                                default_value=priors[hp.name],
                                log=hp.log
                            )
                        )
                    case _:
                        raise ValueError(
                            f"Unknown distribution type {distribution} for {hp.name}."
                        )
            case IntegerHyperparameter():
                match distribution:
                    case "uniform":
                        cs.add(
                            UniformIntegerHyperparameter(
                                name=hp.name,
                                lower=hp.lower,
                                upper=hp.upper,
                                default_value=priors[hp.name],
                                log=hp.log
                            )
                        )
                    case "normal":
                        cs.add(
                            NormalIntegerHyperparameter(
                                name=hp.name,
                                mu=priors[hp.name],
                                sigma=sigma,
                                lower=hp.lower,
                                upper=hp.upper,
                                default_value=priors[hp.name],
                                log=hp.log
                            )
                        )
                    case "beta":
                        cs.add(
                            BetaIntegerHyperparameter(
                                name=hp.name,
                                alpha=alpha,
                                beta=beta,
                                lower=hp.lower,
                                upper=hp.upper,
                                default_value=priors[hp.name],
                                log=hp.log
                            )
                        )
                    case _:
                        raise ValueError(
                            f"Unknown distribution type {distribution} for {hp.name}."
                        )
            case CategoricalHyperparameter():
                cs.add(
                    CategoricalHyperparameter(
                        name=hp.name,
                        choices=hp.choices,
                        default_value=priors[hp.name]
                    )
                )
            case OrdinalHyperparameter():
                cs.add(
                    OrdinalHyperparameter(
                        name=hp.name,
                        sequence=hp.sequence,
                        default_value=priors[hp.name]
                    )
                )
            case Constant():
                cs.add(hp)
            case _:
                raise ValueError(
                    f"Unknown hyperparameter type {type(hp).__name__} in config space."
                )
    return cs


from __future__ import annotations

import logging
import warnings
from collections.abc import Iterator
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from hpoglue import BenchmarkDescription, Measure, TabularBenchmark
from hpoglue.env import Env
from hpoglue.fidelity import RangeFidelity

from hposuite.constants import DATA_DIR
from hposuite.utils import is_package_installed

if TYPE_CHECKING:
    from hpoglue import Config

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)


url: str = "https://figshare.com/ndownloader/files/21188607"


def _download_lcbench_tabular(datadir: Path) -> None:
    import shutil
    import urllib.request
    import zipfile

    datadir.mkdir(parents=True, exist_ok=True)
    zippath = datadir / "data_2k.zip"
    if not zippath.exists():
        _urlopen = urllib.request.urlopen
        print("Downloading from {url}")     # noqa: T201
        with _urlopen(url) as response, zippath.open("wb") as f:
            shutil.copyfileobj(response, f)

    with zipfile.ZipFile(zippath, "r") as zip_ref:
        zip_ref.extractall(datadir)

    print("Downloaded lcbench_tabular data to {datadir.resolve()}")     # noqa: T201


def _process(data_dir: Path) -> None:
    import json

    filepath = data_dir / "data_2k.json"
    print("Processing lcbench_tabular ...")     # noqa: T201
    with filepath.open("r") as f:
        all_data = json.load(f)

    for dataset_name, data in all_data.items():
        logger.info(f"Processing {dataset_name}")
        config_frames_for_dataset = []
        for config_id, config_data in data.items():
            config: dict = config_data["config"]

            log_data: dict = config_data["log"]
            loss: list[str] = log_data["Train/loss"]  # Name of the loss
            val_ce: list[float] = log_data["Train/val_cross_entropy"]
            val_acc: list[float] = log_data["Train/val_accuracy"]
            val_bal_acc: list[float] = log_data["Train/val_balanced_accuracy"]

            test_ce: list[float] = log_data["Train/test_cross_entropy"]
            test_bal_acc: list[float] = log_data["Train/test_balanced_accuracy"]

            # NOTE: Due to there being a lack of "Test/val_accuracy" in the
            # data but a "Train/test_result" we use the latter as the test accuracy
            test_acc: list[float] = log_data["Train/test_result"]

            time = log_data["time"]
            epoch = log_data["epoch"]

            _df = pd.DataFrame(
                {
                    "time": time,
                    "epoch": epoch,
                    "loss": loss,
                    "val_accuracy": val_acc,
                    "val_cross_entropy": val_ce,
                    "val_balanced_accuracy": val_bal_acc,
                    "test_accuracy": test_acc,
                    "test_cross_entropy": test_ce,
                    "test_balanced_accuracy": test_bal_acc,
                },
            )
            # These are single valued but this will make them as a list into
            # the dataframe
            _df = _df.assign(**{"id": config_id, **config})

            config_frames_for_dataset.append(_df)

        #                     | **metrics, **config_params
        # (config_id, epoch)  |

        df_for_dataset = (
            pd.concat(config_frames_for_dataset, ignore_index=True)
            .convert_dtypes()
            .set_index(["id", "epoch"])
            .sort_index()
        )
        table_path = data_dir / f"{dataset_name}.parquet"
        df_for_dataset.to_parquet(table_path)
        logger.info(f"Processed {dataset_name} to {table_path}")


def _setup_lcbench_tabular(datadir: Path) -> None:
    datadir = datadir / "lcbench_tabular"
    if datadir.exists():
        print(f"Data directory {datadir.resolve()} already exists")     # noqa: T201
        return
    _download_lcbench_tabular(datadir)
    _process(datadir)


lcbench_config_keys = [
    "batch_size",
    "max_dropout",
    "max_units",
    "num_layers",
    "learning_rate",
    "momentum",
    "weight_decay",

    # Optional keys
    # "loss",
    # "imputation_strategy",
    # "learning_rate_scheduler",
    # "network",
    # "normalization_strategy",
    # "optimizer",
    # "cosine_annealing",
    # "cosine_annealing_eta_min",
    # "activation",
    # "mlp_shape",
]


def _get_lcbench_space(task_id: str, datadir: Path) -> list[Config]:
    table = _load_lcbench_table(task_id, datadir)
    return TabularBenchmark.get_tabular_config_space(table, lcbench_config_keys)


def _load_lcbench_table(task_id: str, datadir: Path) -> pd.DataFrame:
    return pd.read_parquet(datadir / f"{task_id}.parquet")


def _get_lcbench_tabular_bench(
    description: BenchmarkDescription,
    *,
    task_id: str,
    datadir: Path,
) -> TabularBenchmark:
    table = _load_lcbench_table(task_id, datadir)

    return TabularBenchmark(
        desc=description,
        table=table,
        id_key="id",  # Key in the table to uniquely identify configs
        config_keys=lcbench_config_keys,  # Keys in the table that correspond to configs
    )


def lcbench_tabular(datadir: Path) -> Iterator[BenchmarkDescription]:
    """Generates benchmark descriptions for the LCBench tabular Benchmark.

    Args:
        datadir (Path | None): The directory where the data is stored.
        If None, the default directory is used.

    Yields:
        Iterator[BenchmarkDescription]: An iterator over BenchmarkDescription objects
        for each task in the LCBench tabular Benchmark.
    """
    task_ids = (
        "adult",
        "airlines",
        "albert",
        "Amazon_employee_access",
        "APSFailure",
        "Australian",
        "bank-marketing",
        "blood-transfusion-service-center",
        "car",
        "christine",
        "cnae-9",
        "connect-4",
        "covertype",
        "credit-g",
        "dionis",
        "fabert",
        "Fashion-MNIST",
        "helena",
        "higgs",
        "jannis",
        "jasmine",
        "jungle_chess_2pcs_raw_endgame_complete",
        "kc1",
        "KDDCup09_appetency",
        "kr-vs-kp",
        "mfeat-factors",
        "MiniBooNE",
        "nomao",
        "numerai28.6",
        "phoneme",
        "segment",
        "shuttle",
        "sylvine",
        "vehicle",
        "volkert",
    )

    env = Env(
        name="py310-lcbench-tabular",
        python_version="3.10",
        requirements=(
            "pandas",
            "pyarrow"
        ),
        post_install=_setup_cmd(datadir),
    )
    for req in env.requirements:
        if not is_package_installed(req):
            logger.warning(
                f"Please install the required package for lcbench_tabular: {req}",
                stacklevel=2
            )
            return
    for task_id in task_ids:
        try:
            space = _get_lcbench_space(
                task_id=task_id,
                datadir=datadir
            )
        except FileNotFoundError:
            logger.error(
                f"lcbench_tabular data not found in Data directory {datadir.resolve()}. "
                "Run `python -m hposuite.benchmarks.lcbench_tabular setup` "
                "to download and process the benchmark data."
            )
            return
        yield BenchmarkDescription(
            name=f"lcbench_tabular-{task_id}",
            config_space=space,
            load=partial(_get_lcbench_tabular_bench, task_id=task_id, datadir=datadir),
            is_tabular=True,
            fidelities={
                "epoch": RangeFidelity.from_tuple((1, 51, 1), supports_continuation=True),
            },
            costs={
                "time": Measure.cost((0.0, np.inf), minimize=True),
            },
            metrics={
                "val_accuracy": Measure.metric((0.0, 100.0), minimize=False),
                "val_cross_entropy": Measure.metric((0.0, np.inf), minimize=True),
                "val_balanced_accuracy": Measure.metric((0.0, 1.0), minimize=False),
            },
            test_metrics={
                "test_accuracy": Measure.metric((0.0, 100.0), minimize=False),
                "test_balanced_accuracy": Measure.test_metric((0.0, 1.0), minimize=False),
                "test_cross_entropy": Measure.test_metric(bounds=(0.0, np.inf), minimize=True),
            },
            env=env,
            mem_req_mb=4096,
        )


def _setup_cmd(datadir: Path) -> tuple[str, ...]:
    install_cmd = f"python -m hposuite.benchmarks.lcbench_tabular setup --data_dir {datadir}"
    if datadir is not None:
        install_cmd += f" --data-dir {datadir.resolve()}"
    return tuple(install_cmd.split(" "))


def lcbench_tabular_benchmarks(datadir: Path | None = None) -> Iterator[BenchmarkDescription]:
    """Generator function that yields benchmark descriptions from the lcbench_tabular benchmark."""
    if isinstance(datadir, str):
        datadir = Path(datadir).resolve() / "lcbench_tabular"
    elif datadir is None:
        datadir = DATA_DIR / "lcbench_tabular"
    yield from lcbench_tabular(datadir=datadir)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Setup LCBench Tabular Benchmark data")
    subparsers = parser.add_subparsers(dest="command")
    setup_parser = subparsers.add_parser(
        "setup",
        help="Download and process the LCBench Tabular Benchmark data",
    )
    setup_parser.add_argument(
        "--datadir",
        type=str,
        default=DATA_DIR,
        help="Absolute path of the directory to store the lcbench_tabular data",
    )

    args = parser.parse_args()
    if isinstance(args.datadir, str):
        args.datadir = Path(args.datadir).absolute().resolve()
    _setup_lcbench_tabular(args.datadir)


"""hposuite interface for the DEHB Optimizer."""

from __future__ import annotations

from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any

from ConfigSpace import ConfigurationSpace
from dehb import DEHB
from hpoglue import Config, Optimizer, Problem, Query
from hpoglue.env import Env

if TYPE_CHECKING:
    from hpoglue import Result


class DEHB_Optimizer(Optimizer):
    """The DEHB Optimizer."""

    name = "DEHB"

    env = Env(
        name="dehb-0.1.2",
        python_version="3.10",
        requirements=("dehb==0.1.2"),
    )

    support = Problem.Support(
        fidelities=("single"),
        objectives=("single",),
        cost_awareness=(None, "single"),
        tabular=False,
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        eta: int = 3,
        verbose: bool = False,
        # TODO(eddiebergman): Add more DEHB parameters
    ):
        """Create a DEHB Optimizer instance for a given problem.

        Args:
            problem: The problem to optimize.

            seed: The random seed to use.

            working_directory: The directory to store DEHB's output.

            eta: Sets the aggressiveness of Hyperband's aggressive
                early stopping by retaining 1/eta configurations every round.
                Defaults to 3.

            verbose: Whether to enable verbose logging
        """
        config_space = problem.config_space
        match config_space:
            case ConfigurationSpace():
                pass
            case list():
                raise NotImplementedError("# TODO: Tabular not yet implemented for DEHB!")
            case _:
                raise TypeError("Config space must be a list or a ConfigurationSpace!")

        match problem.fidelities:
            case None:
                min_fidelity = None
                max_fidelity = None
            case (_, fidelity):
                min_fidelity = fidelity.min
                max_fidelity = fidelity.max
            case Mapping():
                raise NotImplementedError("# TODO: Manyfidelity not yet implemented for DEHB!")
            case _:
                raise TypeError("Fidelity must be a string or a list of strings!")

        working_directory.mkdir(parents=True, exist_ok=True)

        self.problem = problem

        # TODO(eddiebergman): Clarify if DEHB is actually cost-aware in
        # terms of how it optimizes or does it just track cost for the
        # sake of `run()`? We only use `ask()` and `tell()` but it seems
        # to require cost in `tell()`.
        self.dehb = DEHB(
            cs=config_space,
            min_fidelity=min_fidelity,
            max_fidelity=max_fidelity,
            verbose=verbose,
            seed=seed,
            eta=eta,
            n_workers=1,
            output_path=working_directory,
        )

        # HACK(eddiebergman): DEHB doesn't have an option to disable logging
        if not verbose:
            self.dehb.logger.disable("dehb.optimizers.dehb")

        self._info_lookup: dict[str, dict[str, Any]] = {}


    def ask(self) -> Query:
        """Ask DEHB for a new config to evaluate."""
        info = self.dehb.ask()
        assert isinstance(info, dict)

        match self.problem.fidelities:
            case None:
                fidelity = None
            case (key, fidelity_def):
                _val = info["fidelity"]
                value = int(_val) if fidelity_def.kind is int else _val
                fidelity = (key, value)
            case Mapping():
                raise NotImplementedError("# TODO: many-fidleity not yet implemented for DEHB!")
            case _:
                raise TypeError("Fidelity must be None, a tuple, or a mapping!")

        config_id = info["config_id"]
        raw_config = info["config"]
        name = f"trial_{config_id}"

        return Query(
            config=Config(config_id=name, values=raw_config),
            fidelity=fidelity,
            optimizer_info=info,
        )


    def tell(self, result: Result) -> None:
        """Tell DEHB the result of the query."""
        fitness: float
        match self.problem.objectives:
            case (name, metric):
                fitness = metric.as_minimize(result.values[name])
            case Mapping():
                raise NotImplementedError("# TODO: Multiobjective not yet implemented for DEHB!")
            case _:
                raise TypeError("Objective must be a string or a list of strings!")

        match self.problem.costs:
            case None:
                cost = None
            case tuple():
                raise NotImplementedError("# TODO: Cost-aware not yet implemented for DEHB!")
            case Mapping():
                raise NotImplementedError("# TODO: Cost-aware not yet implemented for DEHB!")
            case _:
                raise TypeError("Cost must be None or a mapping!")

        assert isinstance(result.query.optimizer_info, dict)
        if cost is None:
            self.dehb.tell(
                result.query.optimizer_info,
                # NOTE(eddiebergman): DEHB requires a cost value, even though
                # we never specify anything related to cost.
                {"fitness": fitness, "cost": result.budget_cost},
            )
        else:
            raise NotImplementedError("# TODO: Cost-aware not yet implemented for DEHB!")



from __future__ import annotations

from hpoglue import BenchmarkDescription, Optimizer

from hposuite.benchmarks import BENCHMARKS
from hposuite.optimizers import OPTIMIZERS


def register(*things: type[Optimizer] | BenchmarkDescription) -> None:
    """Register a benchmark or optimizer."""
    for thing in things:
        if isinstance(thing, BenchmarkDescription):
            if thing.name in BENCHMARKS:
                raise ValueError(f"Duplicate benchmark name: {thing.name}")
            BENCHMARKS[thing.name] = thing
        elif issubclass(thing, Optimizer):
            if thing.name in OPTIMIZERS:
                raise ValueError(f"Duplicate optimizer name: {thing.name}")
            OPTIMIZERS[thing.name] = thing
        else:
            raise TypeError(f"Cannot register {thing}!")



from __future__ import annotations

import argparse
from pathlib import Path

from hposuite.study import Study, create_study


def _study_from_yaml_config(yaml_config: Path) -> Study:
    return Study.from_yaml(yaml_config)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--study_name", "-name",
        type=str,
        help="Study name",
    )
    parser.add_argument(
        "--output_dir", "-d",
        type=Path,
        help="Results directory",
    )
    parser.add_argument(
        "--study_config", "-cfg",
        type=Path,
        help="Absolute path to the Study configuration file",
    )
    parser.add_argument(
        "--optimizers", "-o",
        nargs="+",
        type=str,
        help="Optimizer to use",
    )
    parser.add_argument(
        "--benchmarks", "-b",
        nargs="+",
        type=str,
        help="Benchmark to use",
    )
    parser.add_argument(
        "--seeds", "-s",
        nargs="+",
        type=int,
        default=None,
        help="Seed(s) to use",
    )
    parser.add_argument(
        "--num_seeds", "-n",
        type=int,
        default=1,
        help="Number of seeds to be generated. "
        "Only used if seeds is not provided",
    )
    parser.add_argument(
        "--budget", "-bud",
        type=int,
        default=50,
        help="Budget to use",
    )
    parser.add_argument(
        "--overwrite", "-ow",
        action="store_true",
        help="Overwrite existing results",
    )
    parser.add_argument(
        "--no_continuations", "-no_cont",
        action="store_true",
        help="Do not use continuations",
    )
    parser.add_argument(
        "--exec_type", "-x",
        type=str,
        default="sequential",
        choices=["sequential", "parallel"],
        help="Execution type",
    )
    parser.add_argument(
        "--group_by", "-g",
        type=str,
        default=None,
        choices=["opt", "bench", "opt_bench", "seed", "mem"],
        help="Runs dump group by\n"
        "Only used if exec_type is dump"
    )
    parser.add_argument(
        "--on_error", "-oe",
        type=str,
        default="warn",
        choices=["warn", "raise", "ignore"],
        help="Action to take on error",
    )
    parser.add_argument(
        "--auto_env_handling", "-auto_env",
        action="store_true",
        help="Automatically create and use isolated run environments",
    )
    parser.add_argument(
        "--use_continuations_as_budget", "-cont_budget",
        action="store_true",
        help="Use continuations cost as the budget"
    )
    args = parser.parse_args()

    if args.study_config:
        study = _study_from_yaml_config(args.study_config)
    else:
        study = create_study(
            output_dir=args.output_dir,
            name=args.study_name,
            optimizers=args.optimizers,
            benchmarks=args.benchmarks,
            seeds=args.seeds,
            num_seeds=args.num_seeds,
            budget=args.budget,
            group_by=args.group_by,
            on_error=args.on_error,
        )
    study.optimize(
        continuations=not args.no_continuations,
        overwrite=args.overwrite,
        exec_type=args.exec_type,
        auto_env_handling=args.auto_env_handling,
        use_continuations_as_budget=args.use_continuations_as_budget,
    )




from __future__ import annotations

from itertools import cycle

import matplotlib.pyplot as plt
from more_itertools import take

MARKERS = [
    "o",
    "X",
    "^",
    "H",
    ">",
    "^",
    "p",
    "P",
    "*",
    "h",
    "<",
    "s",
    "x",
    "+",
    "D",
    "d",
    "|",
    "_",
]


def categorical_colors(n: int) -> list[str]:
    """Get a categorical color palette."""
    if n <= 10:  # noqa: PLR2004
        return take(n, iter(plt.get_cmap("tab10").colors))  # type: ignore

    return take(n, cycle(plt.get_cmap("tab20").colors))  # type: ignore


def distinct_markers(n: int) -> list[str]:
    """Get a distinct set of markers."""
    c = cycle(MARKERS)
    return take(n, c)



from __future__ import annotations

import logging
from functools import partial
from itertools import product
from pathlib import Path

import numpy as np
import pandas as pd
from hpoglue import BenchmarkDescription, Config, Measure, TabularBenchmark

from hposuite.constants import DATA_DIR

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

bbob_functions_dict = {
    #Separable

    "f1": "Sphere",
    "f2": "Ellipsoid separable",
    "f3": "Rastrigin separable",
    "f4": "Skew Rastrigin-Bueche",
    "f5": "Linear slope",

    # Low or moderate conditioning
    "f6": "Attractive sector",
    "f7": "Step-ellipsoid",
    "f8": "Rosenbrock original",
    "f9": "Rosenbrock rotated",

    # High conditioning and unimodal
    "f10": "Ellipsoid",
    "f11": "Discus",
    "f12": "Bent cigar",
    "f13": "Sharp ridge",
    "f14": "Sum of different powers",

    # Multi-modal with adequate global structure
    "f15": "Rastrigin",
    "f16": "Weierstrass",
    "f17": "Schaffer F7, condition 10",
    "f18": "Schaffer F7, condition 1000",
    "f19": "Griewank-Rosenbrock F8F2",

    # Multi-modal with weak global structure
    "f20": "Schwefel x*sin(x)",
    "f21": "Gallagher 101 peaks",
    "f22": "Gallagher 21 peaks",
    "f23": "Katsuura",
    "f24": "Lunacek bi-Rastrigin"
}


bbob_dims = (2, 3, 5, 10, 20, 40)
bbob_instances = (0, 1, 2)

def _get_bbob_space(
    function_name: str,
    datadir: str | Path | None = None,
) -> list[Config]:
    """Returns a list of all possible configurations for the BBOB function's Tabular Benchmark."""
    table = _get_bbob_table(function_name, datadir)
    config_keys = [k for k in table.columns if "x" in k]
    return TabularBenchmark.get_tabular_config_space(table, config_keys)


def _get_bbob_table(
    function_name: str,
    datadir: str | Path | None = None,
) -> pd.DataFrame:
    return pd.read_parquet(datadir / f"{function_name}.parquet")


def _get_bbob_tabular_benchmark(
    description: BenchmarkDescription,
    *,
    function_name: str,
    datadir: str | Path | None = None,
) -> TabularBenchmark:
    """Creates a TabularBenchmark object for the BBOB Tabular Benchmark Suite."""
    try:
        bench = _get_bbob_table(function_name, datadir)
    except FileNotFoundError as e:
        logger.error(
            f"Could not find BBOB Tabular Benchmark data for {function_name}. Skipping. "
            f"Run `python -m hposuite.benchmarks.create_tabular "
            f"--benchmark bbob-{function_name} -suite bbob_tabular --task {function_name}`"
            "to create the benchmark data."
        )
        raise e
    config_keys = [k for k in bench.columns if "x" in k]
    return TabularBenchmark(
        desc=description,
        table=bench,
        id_key="config_id",
        config_keys=config_keys,
    )


def bbob_tabular_desc(datadir: str | Path | None = None) -> BenchmarkDescription:
    """Generates benchmark descriptions for the Synthetic BBOB Tabular Benchmark suite.

    This function iterates over all combinations of function IDs, dimensions, and instances
    defined in `bbob_functions_dict`, `bbob_dims`, and `bbob_instances` respectively. For each
    combination, it creates a `BenchmarkDescription` object with the corresponding configuration
    space, benchmark function, and environment settings.
    """
    for f_id, dim, ins in product(bbob_functions_dict.keys(), bbob_dims, bbob_instances):
        function_name = f"{f_id}-{dim}-{ins}"
        try:
            space = _get_bbob_space(function_name, datadir)
        except FileNotFoundError:
            continue
        yield BenchmarkDescription(
            name=f"bbob_tabular-{function_name}",
            config_space=space,
            load=partial(
                _get_bbob_tabular_benchmark,
                function_name=function_name,
                datadir=datadir
            ),
            metrics={
                "value": Measure.metric((-np.inf, np.inf), minimize=True),
            },
            test_metrics=None,
            costs=None,
            fidelities=None,
            has_conditionals=False,
            is_tabular=True,
            mem_req_mb=1024,
        )


def bbob_tabular_benchmarks(datadir: str | Path | None = None) :
    """A generator that yields all BBOB Tabular benchmarks."""
    if isinstance(datadir, str):
        datadir = Path(datadir).resolve() / "bbob_tabular"
    elif datadir is None:
        datadir = DATA_DIR / "bbob_tabular"

    yield from bbob_tabular_desc(datadir=datadir)


"""Interfacing benchmarks from the Pymoo library."""

from __future__ import annotations

from collections.abc import Iterator
from functools import partial
from typing import TYPE_CHECKING, Any

import numpy as np
from ConfigSpace import ConfigurationSpace, Float
from hpoglue import BenchmarkDescription, Measure, Result, SurrogateBenchmark
from hpoglue.env import Env

if TYPE_CHECKING:
    import pymoo
    from hpoglue import Query

def get_pymoo_space(
    pymoo_prob: pymoo.core.problem.Problem
) -> ConfigurationSpace:
    """Get ConfigSpace from pymoo problem."""
    n_var = pymoo_prob.n_var
    xl, xu = pymoo_prob.xl, pymoo_prob.xu
    return ConfigurationSpace(
        {f"x{i}": (xl[i], xu[i]) for i in range(n_var)}
    )


def _pymoo_query_function(
    query: Query,
    benchmark: pymoo.core.problem.Problem,
) -> Result:
    assert query.fidelity is None
    config_vals = np.array(list(query.config.values.values()))
    values = benchmark.evaluate(config_vals).tolist()
    if len(values) > 1:
        values = {f"value{i}": val for i, val in enumerate(values, start=1)}
    else:
        values = {"value": values[0]}
    return Result(
        query=query,
        values=values,
        fidelity=None,
    )


def _get_pymoo_problems(
    function_name: str,
    **kwargs: Any,
)-> pymoo.core.problem.Problem:

    import pymoo.problems

    match function_name:
        case "omnitest":
            from pymoo.problems.multi.omnitest import OmniTest
            return OmniTest()
        case "sympart":
            from pymoo.problems.multi.sympart import SYMPART
            return SYMPART()
        case "sympart_rotated":
            from pymoo.problems.multi.sympart import SYMPARTRotated
            return SYMPARTRotated()
        case _:
            return pymoo.problems.get_problem(function_name, **kwargs)


def _pymoo_surrogate_bench(
    desc: BenchmarkDescription,
    pymoo_prob: pymoo.core.problem.Problem,
) -> SurrogateBenchmark:
    query_function = partial(_pymoo_query_function, benchmark=pymoo_prob)
    return SurrogateBenchmark(
        desc=desc,
        config_space=get_pymoo_space(pymoo_prob),
        benchmark=pymoo_prob,
        query=query_function,
    )


_pymoo_so = [
    "ackley",
    "griewank",
    "himmelblau",
    "rastrigin",
    "rosenbrock",
    "schwefel",
    "sphere",
    "zakharov",
]

def pymoo_so_problems() -> Iterator[BenchmarkDescription]:
    """Generates benchmark descriptions for single-objective problems using the pymoo library."""
    env = Env(
        name="py310-pymoo-0.6.1.3",
        python_version="3.10",
        requirements=("pymoo==0.6.1.3"),
        post_install=None
    )
    for prob_name in _pymoo_so:
        pymoo_prob = _get_pymoo_problems(prob_name)
        yield BenchmarkDescription(
            name=f"pymoo-{prob_name}",
            config_space=get_pymoo_space(pymoo_prob),
            load = partial(_pymoo_surrogate_bench, pymoo_prob=pymoo_prob),
            has_conditionals=False,
            metrics={
                    "value": Measure.metric((-np.inf, np.inf), minimize=True),
                },
            is_tabular=False,
            env=env,
            mem_req_mb=1024,
        )

_pymoo_mo = [
    "kursawe",
    "zdt1",
    "zdt2",
    "zdt3",
    "zdt4",
    "zdt5",
    "zdt6",
    "omnitest",
    "sympart",
    "sympart_rotated",
]

def pymoo_mo_problems() -> Iterator[BenchmarkDescription]:
    """Generator function that yields benchmark descriptions for multi-objective problems
    using the pymoo library.
    """
    env = Env(
        name="py310-pymoo-0.6.1.3",
        requirements=("pymoo==0.6.1.3"),
        post_install=None
    )
    for prob_name in _pymoo_mo:
        pymoo_prob = _get_pymoo_problems(prob_name)
        yield BenchmarkDescription(
            name=f"pymoo-{prob_name}",
            config_space=get_pymoo_space(pymoo_prob),
            load = partial(_pymoo_surrogate_bench,  pymoo_prob=pymoo_prob),
            has_conditionals=False,
            metrics={
                    "value1": Measure.metric((-np.inf, np.inf), minimize=True),
                    "value2": Measure.metric((-np.inf, np.inf), minimize=True),
                },
            is_tabular=False,
            env=env,
            mem_req_mb=1024,
        )


_pymoo_many_obj = [
    "dtlz1",
    "dtlz2",
    "dtlz3",
    "dtlz4",
    "dtlz5",
    "dtlz6",
    "dtlz7",
]


def pymoo_many_obj_problems() -> Iterator[BenchmarkDescription]:
    """Generates benchmark descriptions for many-objective problems using the PyMOO library."""
    env = Env(
        name="py310-pymoo-0.6.1.3",
        requirements=("pymoo==0.6.1.3"),
        post_install=None
    )
    for prob_name in _pymoo_many_obj:
        pymoo_prob = _get_pymoo_problems(prob_name)
        yield BenchmarkDescription(
            name=f"pymoo-{prob_name}",
            config_space=get_pymoo_space(pymoo_prob),

            load = partial(_pymoo_surrogate_bench,  pymoo_prob=pymoo_prob),
            has_conditionals=False,
            # PyMOO DTLZ problems have 3 objectives by default, can be changed by setting n_obj
            metrics={
                    "value1": Measure.metric((-np.inf, np.inf), minimize=True),
                    "value2": Measure.metric((-np.inf, np.inf), minimize=True),
                    "value3": Measure.metric((-np.inf, np.inf), minimize=True),
                },
            is_tabular=False,
            env=env,
            mem_req_mb=1024,
        )


def pymoo_benchmarks() -> Iterator[BenchmarkDescription]:
    """Generate benchmark descriptions for various pymoo problems.

    This function yields benchmark descriptions for single-objective,
    multi-objective, and many-objective problems from the pymoo library.
    """
    yield from pymoo_so_problems()
    yield from pymoo_mo_problems()
    yield from pymoo_many_obj_problems()



from __future__ import annotations

import logging
import warnings
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import neps
import numpy as np
from hpoglue import Config, Optimizer, Problem, Query, Result
from hpoglue.env import Env
from neps import AskAndTell, algorithms

from hposuite.utils import set_priors_as_defaults

if TYPE_CHECKING:
    from ConfigSpace import ConfigurationSpace
    from hpoglue.fidelity import Fidelity

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def set_seed(seed: int) -> None:
    """Set the seed for the optimizer."""
    import random

    import numpy as np
    import torch

    torch.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)  # noqa: NPY002


class NepsOptimizer(Optimizer):
    """Base class for Neps Optimizers."""
    name = "NepsOptimizer"


    def __init__(
        self,
        *,
        problem: Problem,
        space: neps.SearchSpace,
        optimizer: str,
        seed: int,
        working_directory: str | Path,
        fidelities: tuple[str, Fidelity] | None = None,
        random_weighted_opt: bool = False,
        scalarization_weights: Literal["equal", "random"] | Mapping[str, float] = "random",
        **kwargs: Any,
    ) -> None:
        """Initialize the optimizer."""
        self.problem = problem
        self.space = space

        match fidelities:
            case None:
                pass
            case (fid_name, fidelity):
                _fid = fidelity
                min_fidelity = fidelity.min
                max_fidelity = fidelity.max
                match _fid.kind:
                    case _ if _fid.kind is int:
                        space.fidelities = {
                            f"{fid_name}": neps.Integer(
                                lower=min_fidelity, upper=max_fidelity, is_fidelity=True
                            )
                        }
                    case _ if _fid.kind is float:
                        space.fidelities = {
                            f"{fid_name}": neps.Float(
                                lower=min_fidelity, upper=max_fidelity, is_fidelity=True
                            )
                        }
                    case _:
                        raise TypeError(
                            f"Invalid fidelity type: {type(_fid.kind).__name__}. "
                            "Expected int or float."
                        )
            case _:
                raise TypeError("Fidelity must be a tuple or None.")


        if not isinstance(working_directory, Path):
            working_directory = Path(working_directory)
        self.seed = seed
        self.working_dir = working_directory

        self.optimizer = AskAndTell(
            algorithms.PredefinedOptimizers[optimizer](
                space,
                **kwargs,
            )
        )
        self.trial_counter = 0

        self.objectives = self.problem.get_objectives()
        self.random_weighted_opt = random_weighted_opt
        self.scalarization_weights = None

        if self.random_weighted_opt:
            assert len(self.objectives) > 1, (
                "Random weighted optimization is only supported for multi-objective problems."
            )
            match scalarization_weights:
                case Mapping():
                    self.scalarization_weights = scalarization_weights
                case "equal":
                    self.scalarization_weights = (
                        dict.fromkeys(self.objectives, 1.0 / len(self.objectives))
                    )
                case "random":
                    weights = np.random.uniform(size=len(self.objectives))  # noqa: NPY002
                    self.scalarization_weights = dict(zip(self.objectives, weights, strict=True))
                case _:
                    raise ValueError(
                        f"Invalid scalarization_weights: {scalarization_weights}. "
                        "Expected 'equal', 'random', or a Mapping."
                    )


    def ask(self) -> Query:
        """Ask the optimizer for a new trial."""
        import copy
        trial = self.optimizer.ask() # TODO: Figure out fidelity
        fidelity = None
        _config = copy.deepcopy(trial.config)
        match self.problem.fidelities:
            case None:
                pass
            case Mapping():
                raise NotImplementedError("Many-fidelity not yet implemented for NePS.")
            case (fid_name, _):
                _fid_value = _config.pop(fid_name)
                fidelity = (fid_name, _fid_value)
            case _:
                raise TypeError(
                    "Fidelity must be a tuple or a Mapping. \n"
                    f"Got {type(self.problem.fidelities)}."
                )

        self.trial_counter += 1
        return Query(
            config = Config(config_id=self.trial_counter, values=_config),
            fidelity=fidelity,
            optimizer_info=trial
        )


    def tell(self, result: Result) -> None:
        """Tell the optimizer about the result of a trial."""
        match self.problem.objectives:
            case (name, metric):
                _values = result.values[name]
                # Normalize objective value for NepsIFBO
                if self.name == "NepsIFBO":
                    bounds_min = metric.bounds[0]
                    bounds_max = metric.bounds[1]
                    _values = (_values - bounds_min) / (bounds_max - bounds_min)
                _values = metric.as_minimize(_values)
            case Mapping():
                _values = {
                    key: obj.as_minimize(result.values[key])
                    for key, obj in self.problem.objectives.items()
                }
                if self.random_weighted_opt:
                    _values = sum(
                        self.scalarization_weights[obj] * _values[obj] for obj in self.objectives
                    )
                else:
                    _values = list(_values.values())
            case _:
                raise TypeError(
                    "Objective must be a tuple or a Mapping! "
                    f"Got {type(self.problem.objectives)}."
                )

        match self.problem.costs:
            case None:
                pass
            case tuple():
                raise NotImplementedError("# TODO: Cost-aware not yet implemented for NePS!")
            case Mapping():
                raise NotImplementedError("# TODO: Cost-aware not yet implemented for NePS!")
            case _:
                raise TypeError("Cost must be None or a mapping!")

        self.optimizer.tell(
            trial=result.query.optimizer_info,
            result=_values,
        )


class NepsBO(NepsOptimizer):
    """Bayesian Optimization in Neps."""

    name = "NepsBO"

    support = Problem.Support(
        fidelities=(None,),
        objectives=("single",),
        cost_awareness=(None,),
        tabular=False,
    )

    env = Env(
        name="Neps-0.13.0",
        python_version="3.10",
        requirements=("neural-pipeline-search>=0.13.0",)
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: str | Path,
        initial_design_size: int | Literal["ndim"] = "ndim",
    ) -> None:
        """Initialize the optimizer."""
        match problem.fidelities:
            case None:
                pass
            case Mapping() | tuple():
                raise ValueError("NepsBO does not support fidelities.")
            case _:
                raise TypeError("Fidelity must be a tuple or a Mapping.")

        match problem.objectives:
            case tuple():
                pass
            case Mapping():
                raise ValueError("NepsBO only supports single-objective problems.")
            case _:
                raise TypeError(
                    "Objectives must be a tuple or a Mapping. \n"
                    f"Got {type(problem.objectives)}."
                )

        space = configspace_to_pipeline_space(problem.config_space)
        set_seed(seed)

        super().__init__(
            problem=problem,
            space=space,
            seed=seed,
            working_directory=working_directory,
            optimizer="bayesian_optimization",
            initial_design_size=initial_design_size,
        )


class NepsRW(NepsOptimizer):
    """Random Weighted Scalarization of objectives using Bayesian Optimization in Neps."""

    name = "NepsRW"

    support = Problem.Support(
        fidelities=(None,),
        objectives=("many"),
        cost_awareness=(None,),
        tabular=False,
    )

    env = Env(
        name="Neps-0.13.0",
        python_version="3.10",
        requirements=("neural-pipeline-search>=0.13.0",)
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: str | Path,
        scalarization_weights: Literal["equal", "random"] | Mapping[str, float] = "random",
        initial_design_size: int | Literal["ndim"] = "ndim",
    ) -> None:
        """Initialize the optimizer."""
        space = configspace_to_pipeline_space(problem.config_space)


        match problem.fidelities:
            case None:
                pass
            case Mapping() | tuple():
                raise ValueError("NepsRW does not support fidelities.")
            case _:
                raise TypeError("Fidelity must be a tuple or a Mapping.")

        match problem.objectives:
            case tuple():
                raise ValueError("NepsRW only supports multi-objective problems.")
            case Mapping():
                pass
            case _:
                raise TypeError(
                    "Objectives must be a tuple or a Mapping. \n"
                    f"Got {type(problem.objectives)}."
                )

        set_seed(seed)

        super().__init__(
            problem=problem,
            space=space,
            optimizer="bayesian_optimization",
            seed=seed,
            working_directory=working_directory,
            random_weighted_opt=True,
            scalarization_weights=scalarization_weights,
            initial_design_size=initial_design_size,
        )


class NepsSuccessiveHalving(NepsOptimizer):
    """Neps Successive Halving."""

    name = "NepsSuccessiveHalving"

    support = Problem.Support(
        fidelities=("single",),
        objectives=("single"),
        cost_awareness=(None,),
        tabular=False,
        continuations=True,
    )

    env = Env(
        name="Neps-0.13.0",
        python_version="3.10",
        requirements=("neural-pipeline-search>=0.13.0",)
    )

    mem_req_mb = 1024

    def __init__(
        self,
        problem: Problem,
        seed: int,
        working_directory: str | Path,
        eta: int = 3,
        sampler: Literal["uniform", "prior"] = "uniform",
    ) -> None:
        """Initialize the optimizer."""
        space = configspace_to_pipeline_space(problem.config_space)

        _fid = None
        match problem.fidelities:
            case None:
                raise ValueError("NepsSuccessiveHalving requires a fidelity.")
            case Mapping():
                raise NotImplementedError(
                    "Many-fidelity not yet implemented for NepsSuccessiveHalving."
                )
            case (fid_name, fidelity):
                _fid = (fid_name, fidelity)
            case _:
                raise TypeError("Fidelity must be a tuple or a Mapping.")

        match problem.objectives:
            case tuple():
                pass
            case Mapping():
                raise ValueError("NepsSuccessiveHalving only supports single-objective problems.")
            case _:
                raise TypeError(
                    "Objectives must be a tuple or a Mapping. \n"
                    f"Got {type(problem.objectives)}."
                )

        set_seed(seed)

        super().__init__(
            problem=problem,
            space=space,
            seed=seed,
            working_directory=working_directory,
            optimizer="successive_halving",
            fidelities=_fid,
            eta=eta,
            sampler=sampler,
        )


class NepsHyperband(NepsOptimizer):
    """NepsHyperband."""

    name = "NepsHyperband"

    support = Problem.Support(
        fidelities=("single",),
        objectives=("single"),
        cost_awareness=(None,),
        tabular=False,
        continuations=True,
    )

    env = Env(
        name="Neps-0.13.0",
        python_version="3.10",
        requirements=("neural-pipeline-search>=0.13.0",)
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: str | Path,
        eta: int = 3,
        sampler: Literal["uniform", "prior"] = "uniform",
    ) -> None:
        """Initialize the optimizer."""
        space = configspace_to_pipeline_space(problem.config_space)

        _fid = None
        match problem.fidelities:
            case None:
                raise ValueError("NepsHyperband requires a fidelity.")
            case Mapping():
                raise NotImplementedError("Many-fidelity not yet implemented for NepsHyperband.")
            case (fid_name, fidelity):
                _fid = (fid_name, fidelity)
            case _:
                raise TypeError("Fidelity must be a tuple or a Mapping.")

        match problem.objectives:
            case tuple():
                pass
            case Mapping():
                raise ValueError("NepsHyperband only supports single-objective problems.")
            case _:
                raise TypeError(
                    "Objectives must be a tuple or a Mapping. \n"
                    f"Got {type(problem.objectives)}."
                )

        set_seed(seed)

        super().__init__(
            problem=problem,
            space=space,
            seed=seed,
            working_directory=working_directory,
            optimizer="hyperband",
            fidelities=_fid,
            eta=eta,
            sampler=sampler,
        )

class NepsHyperbandRW(NepsOptimizer):
    """Random Weighted Scalarization of objectives using Hyperband for budget allocation in Neps."""

    name = "NepsHyperbandRW"

    support = Problem.Support(
        fidelities=("single",),
        objectives=("many"),
        cost_awareness=(None,),
        tabular=False,
        continuations=True,
    )

    env = Env(
        name="Neps-0.13.0",
        python_version="3.10",
        requirements=("neural-pipeline-search>=0.13.0",)
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: str | Path,
        scalarization_weights: Literal["equal", "random"] | Mapping[str, float] = "random",
        eta: int = 3,
        sampler: Literal["uniform", "prior"] = "uniform",
    ) -> None:
        """Initialize the optimizer."""
        space = configspace_to_pipeline_space(problem.config_space)

        _fid = None
        match problem.fidelities:
            case None:
                raise ValueError("NepsHyperbandRW requires a fidelity.")
            case Mapping():
                raise NotImplementedError("Many-fidelity not yet implemented for NepsHyperbandRW.")
            case (fid_name, fidelity):
                _fid = (fid_name, fidelity)
            case _:
                raise TypeError("Fidelity must be a tuple or a Mapping.")

        match problem.objectives:
            case tuple():
                raise ValueError("NepsHyperbandRW only supports multi-objective problems.")
            case Mapping():
                pass
            case _:
                raise TypeError(
                    "Objectives must be a tuple or a Mapping. \n"
                    f"Got {type(problem.objectives)}."
                )

        set_seed(seed)

        super().__init__(
            problem=problem,
            space=space,
            optimizer="hyperband",
            seed=seed,
            working_directory=working_directory,
            fidelities=_fid,
            random_weighted_opt=True,
            scalarization_weights=scalarization_weights,
            eta=eta,
            sampler=sampler,
        )



class NepsASHA(NepsOptimizer):
    """NepsASHA."""

    name = "NepsASHA"

    support = Problem.Support(
        fidelities=("single",),
        objectives=("single"),
        cost_awareness=(None,),
        tabular=False,
        continuations=True,
    )

    env = Env(
        name="Neps-0.13.0",
        python_version="3.10",
        requirements=("neural-pipeline-search>=0.13.0",)
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: str | Path,
        eta: int = 3,
        sampler: Literal["uniform", "prior"] = "uniform",
    ) -> None:
        """Initialize the optimizer."""
        space = configspace_to_pipeline_space(problem.config_space)

        _fid = None
        match problem.fidelities:
            case None:
                raise ValueError("NepsASHA requires a fidelity.")
            case Mapping():
                raise NotImplementedError("Many-fidelity not yet implemented for NepsASHA.")
            case (fid_name, fidelity):
                _fid = (fid_name, fidelity)
            case _:
                raise TypeError("Fidelity must be a tuple or a Mapping.")

        match problem.objectives:
            case tuple():
                pass
            case Mapping():
                raise ValueError("NepsASHA only supports single-objective problems.")
            case _:
                raise TypeError(
                    "Objectives must be a tuple or a Mapping. \n"
                    f"Got {type(problem.objectives)}."
                )

        set_seed(seed)

        super().__init__(
            problem=problem,
            space=space,
            seed=seed,
            working_directory=working_directory,
            eta=eta,
            optimizer="asha",
            fidelities=_fid,
            sampler=sampler,
        )


class NepsAsyncHB(NepsOptimizer):
    """Neps Async Hyperband."""

    name = "NepsAsyncHB"

    support = Problem.Support(
        fidelities=("single",),
        objectives=("single"),
        cost_awareness=(None,),
        tabular=False,
        continuations=True,
    )

    env = Env(
        name="Neps-0.13.0",
        python_version="3.10",
        requirements=("neural-pipeline-search>=0.13.0",)
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: str | Path,
        eta: int = 3,
    ) -> None:
        """Initialize the optimizer."""
        space = configspace_to_pipeline_space(problem.config_space)

        _fid = None
        match problem.fidelities:
            case None:
                raise ValueError("NepsAsyncHB requires a fidelity.")
            case Mapping():
                raise NotImplementedError("Many-fidelity not yet implemented for NepsAsyncHB.")
            case (fid_name, fidelity):
                _fid = (fid_name, fidelity)
            case _:
                raise TypeError("Fidelity must be a tuple or a Mapping.")

        match problem.objectives:
            case tuple():
                pass
            case Mapping():
                raise ValueError("NepsAsyncHB only supports single-objective problems.")
            case _:
                raise TypeError(
                    "Objectives must be a tuple or a Mapping. \n"
                    f"Got {type(problem.objectives)}."
                )

        set_seed(seed)

        super().__init__(
            problem=problem,
            space=space,
            seed=seed,
            working_directory=working_directory,
            optimizer="async_hb",
            fidelities=_fid,
            eta=eta,
        )


class NepsPriorband(NepsOptimizer):
    """NepsPriorband."""

    name = "NepsPriorband"

    support = Problem.Support(
        fidelities=("single",),
        objectives=("single"),
        cost_awareness=(None,),
        tabular=False,
        priors=True,
    )

    env = Env(
        name="Neps-0.13.0",
        python_version="3.10",
        requirements=("neural-pipeline-search>=0.13.0",)
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: str | Path,
        eta: int = 3,
        sample_prior_first: bool | Literal["highest_fidelity"] = False,
        base: Literal["successive_halving", "hyperband", "asha", "async_hb"] = "hyperband",
    ) -> None:
        """Initialize the optimizer."""
        assert len(problem.priors[1]) == 1, (
            "NepsPriorband only supports single-objective priors. "
        )
        config_space = set_priors_as_defaults(
            config_space=problem.config_space,
            priors=next(iter(problem.priors[1].values())),
            distribution="normal",
        )
        space = configspace_to_pipeline_space(
            config_space,
            use_priors=True,
        )

        _fid = None
        match problem.fidelities:
            case None:
                raise ValueError("NepsPriorband requires a fidelity.")
            case Mapping():
                raise NotImplementedError(
                    "Many-fidelity not yet implemented for NepsPriorband."
                )
            case (fid_name, fidelity):
                _fid = (fid_name, fidelity)
            case _:
                raise TypeError("Fidelity must be a tuple or a Mapping.")

        match problem.objectives:
            case tuple():
                pass
            case Mapping():
                raise ValueError("NepsPriorband only supports single-objective problems.")
            case _:
                raise TypeError(
                    "Objectives must be a tuple or a Mapping. \n"
                    f"Got {type(problem.objectives)}."
                )

        set_seed(seed)

        super().__init__(
            problem=problem,
            space=space,
            seed=seed,
            working_directory=working_directory,
            optimizer="priorband",
            fidelities=_fid,
            eta=eta,
            base=base,
            sample_prior_first=sample_prior_first,
        )


class NepsPiBO(NepsOptimizer):
    """Neps PiBO - Bayesian Optimization with User Beliefs."""

    name = "NepsPiBO"

    support = Problem.Support(
        fidelities=(None,),
        objectives=("single",),
        cost_awareness=(None,),
        tabular=False,
        priors=True,
    )

    env = Env(
        name="Neps-0.13.0",
        python_version="3.10",
        requirements=("neural-pipeline-search>=0.13.0",)
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: str | Path,
        initial_design_size: int | Literal["ndim"] = "ndim",
        sample_prior_first: bool = False,
    ) -> None:
        """Initialize the optimizer."""
        assert len(problem.priors[1]) == 1, (
            "NepsPiBO only supports single-objective priors. "
        )
        config_space = set_priors_as_defaults(
            config_space=problem.config_space,
            priors=next(iter(problem.priors[1].values())),
            distribution="normal",
        )
        space = configspace_to_pipeline_space(
            config_space,
            use_priors=True,
        )

        match problem.fidelities:
            case None:
                pass
            case Mapping() | tuple():
                raise ValueError("NepsPiBO does not support fidelities.")
            case _:
                raise TypeError("Fidelity must be a tuple or a Mapping.")

        match problem.objectives:
            case tuple():
                pass
            case Mapping():
                raise ValueError("NepsPiBO only supports single-objective problems.")
            case _:
                raise TypeError(
                    "Objectives must be a tuple or a Mapping. \n"
                    f"Got {type(problem.objectives)}."
                )

        set_seed(seed)

        super().__init__(
            problem=problem,
            space=space,
            seed=seed,
            working_directory=working_directory,
            optimizer="pibo",
            initial_design_size=initial_design_size,
            sample_prior_first=sample_prior_first,
        )


class NepsIFBO(NepsOptimizer):
    """In-context Freeze Thaw Bayesian Optimization in Neps."""
    name = "NepsIFBO"

    support = Problem.Support(
        fidelities=("single",),
        objectives=("single",),
        cost_awareness=(None,),
        tabular=False,
        continuations=True,
    )

    env = Env(
        name="Neps-0.13.0",
        python_version="3.10",
        requirements=("neural-pipeline-search>=0.13.0",)
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: str | Path,
        initial_design_size: int | Literal["ndim"] = "ndim",
        use_priors: bool = False,
        step_size: int | float = 1,
        surrogate_path: str | Path | None = None,
        surrogate_version: str = "0.0.1",
    ) -> None:
        """Initialize the optimizer."""
        space = configspace_to_pipeline_space(problem.config_space)

        _fid = None
        match problem.fidelities:
            case None:
                raise ValueError("NepsIFBO requires a fidelity.")
            case Mapping():
                raise NotImplementedError("Many-fidelity not yet implemented for NepsIFBO.")
            case (fid_name, fidelity):
                _fid = (fid_name, fidelity)
            case _:
                raise TypeError("Fidelity must be a tuple or a Mapping.")

        match problem.objectives:
            case tuple():
                pass
            case Mapping():
                raise ValueError("NepsIFBO only supports single-objective problems.")
            case _:
                raise TypeError(
                    "Objectives must be a tuple or a Mapping. \n"
                    f"Got {type(problem.objectives)}."
                )

        set_seed(seed)

        super().__init__(
            problem=problem,
            space=space,
            seed=seed,
            working_directory=working_directory,
            optimizer="ifbo",
            fidelities=_fid,
            initial_design_size=initial_design_size,
            use_priors=use_priors,
            step_size=step_size,
            surrogate_path=surrogate_path,
            surrogate_version=surrogate_version,
        )


class NepsMOASHA(NepsOptimizer):
    """NepsMOASHA."""

    name = "NepsMOASHA"

    support = Problem.Support(
        fidelities=("single",),
        objectives=("many"),
        cost_awareness=(None,),
        tabular=False,
        continuations=True,
    )

    env = Env(
        name="Neps-0.13.0",
        python_version="3.10",
        requirements=("neural-pipeline-search==0.13.0",)
    )

    mem_req_mb = 1024

    def __init__(
        self,
        problem: Problem,
        seed: int,
        working_directory: str | Path,
        mo_selector: Literal["nsga2", "epsnet"] = "epsnet",
        sampler: Literal["uniform", "prior"] = "uniform",
        eta: int = 3,
    ) -> None:
        """Initialize the optimizer."""
        space = configspace_to_pipeline_space(problem.config_space)

        _fid = None
        match problem.fidelities:
            case None:
                raise ValueError("NepsMOASHA requires a fidelity.")
            case Mapping():
                raise NotImplementedError("Many-fidelity not yet implemented for NepsMOASHA.")
            case (fid_name, fidelity):
                _fid = (fid_name, fidelity)
            case _:
                raise TypeError("Fidelity must be a tuple or a Mapping.")

        match problem.objectives:
            case tuple():
                raise ValueError("NepsMOASHA only supports multi-objective problems.")
            case Mapping():
                pass
            case _:
                raise TypeError(
                    "Objectives must be a tuple or a Mapping. \n"
                    f"Got {type(problem.objectives)}."
                )
        set_seed(seed)

        super().__init__(
            problem=problem,
            space=space,
            optimizer="moasha",
            seed=seed,
            working_directory=working_directory,
            fidelities=_fid,
            mo_selector=mo_selector,
            sampler=sampler,
            eta=eta,
        )


class NepsMOHyperband(NepsOptimizer):
    """NepsMOHyperband."""

    name = "NepsMOHyperband"

    support = Problem.Support(
        fidelities=("single",),
        objectives=("many"),
        cost_awareness=(None,),
        tabular=False,
        continuations=True,
    )

    env = Env(
        name="Neps-0.13.0",
        python_version="3.10",
        requirements=("neural-pipeline-search==0.13.0",)
    )

    mem_req_mb = 1024

    def __init__(
        self,
        problem: Problem,
        seed: int,
        working_directory: str | Path,
        mo_selector: Literal["nsga2", "epsnet"] = "epsnet",
        sampler: Literal["uniform", "prior"] = "uniform",
        eta: int = 3,
    ) -> None:
        """Initialize the optimizer."""
        space = configspace_to_pipeline_space(problem.config_space)

        _fid = None
        match problem.fidelities:
            case None:
                raise ValueError("NepsMOHyperband requires a fidelity.")
            case Mapping():
                raise NotImplementedError("Many-fidelity not yet implemented for NepsMOHyperband.")
            case (fid_name, fidelity):
                _fid = (fid_name, fidelity)
            case _:
                raise TypeError("Fidelity must be a tuple or a Mapping.")

        match problem.objectives:
            case tuple():
                raise ValueError("NepsMOHyperband only supports multi-objective problems.")
            case Mapping():
                pass
            case _:
                raise TypeError(
                    "Objectives must be a tuple or a Mapping. \n"
                    f"Got {type(problem.objectives)}."
                )
        set_seed(seed)

        super().__init__(
            problem=problem,
            space=space,
            optimizer="mo_hyperband",
            seed=seed,
            working_directory=working_directory,
            fidelities=_fid,
            mo_selector=mo_selector,
            sampler=sampler,
            eta=eta,
        )


def configspace_to_pipeline_space(  # noqa: C901
    config_space: ConfigurationSpace,
    *,
    use_priors: bool = False,
    std: float = 0.25,
) -> neps.SearchSpace:
    """Convert a ConfigurationSpace to a Neps SearchSpace.

    Args:
        config_space: The ConfigurationSpace to convert.
        use_priors: Whether to use priors for the hyperparameters.
            NePS only supports Normally distributed priors, therefore
            if the hyperparameter is not Normally distributed and `use_priors` is True,
            the default value of the hyperparameter is used as the mean to create a
            Normally distributed prior.
            If `use_priors` is False and the hyperparameter is Normally distributed,
            `use_priors` is ignored and it is converted to a NePS hyperparameter with
            its mean as the prior.
        std: The standard deviation of the prior. Only used if `use_priors` is True and/or
            the hyperparameter is Normally distributed.
            Translated to the confidence of the prior in NePS as 1 - std.
    """
    from ConfigSpace.hyperparameters import (
        BetaFloatHyperparameter,
        BetaIntegerHyperparameter,
        CategoricalHyperparameter,
        Constant,
        NormalFloatHyperparameter,
        NormalIntegerHyperparameter,
        OrdinalHyperparameter,
        UniformFloatHyperparameter,
        UniformIntegerHyperparameter,
    )
    from neps.space.parameters import Parameter
    space: dict[str, Parameter | Constant] = {}
    std_to_confidences: dict[float, str] = {
        0.25: "high",
        0.5: "medium",
        0.75: "low",
    }
    if any(config_space.conditions) or any(config_space.forbidden_clauses):
        raise NotImplementedError(
            "The ConfigurationSpace has conditions or forbidden clauses, "
            "which are not supported by neps."
        )
    for hp in list(config_space.values()):
        match hp:
            case NormalFloatHyperparameter():
                if not use_priors:
                    warnings.warn(
                        "use priors=False but hyperparameter is a NormalFloatHyperparameter, "
                        "which defaults to a Prior in NePS. Ignoring use_priors.",
                        stacklevel=2,
                    )
                space[hp.name] = neps.Float(
                    lower=hp.lower,
                    upper=hp.upper,
                    log=hp.log,
                    prior=hp.mu,
                    prior_confidence=std_to_confidences[hp.sigma],
                )
            case UniformFloatHyperparameter() | BetaFloatHyperparameter():
                space[hp.name] = neps.Float(
                    lower=hp.lower,
                    upper=hp.upper,
                    log=hp.log,
                    prior=hp.default_value if use_priors else None,
                    prior_confidence=std_to_confidences[std]
                )
            case NormalIntegerHyperparameter():
                if not use_priors:
                    warnings.warn(
                        "use priors=False but hyperparameter is a NormalIntegerHyperparameter, "
                        "which defaults to a Prior in NePS. Ignoring use_priors.",
                        stacklevel=2,
                    )
                space[hp.name] = neps.Integer(
                    lower=hp.lower,
                    upper=hp.upper,
                    log=hp.log,
                    prior=hp.mu,
                    prior_confidence=std_to_confidences[hp.sigma],
                )
            case UniformIntegerHyperparameter() | BetaIntegerHyperparameter():
                space[hp.name] = neps.Integer(
                    lower=hp.lower,
                    upper=hp.upper,
                    log=hp.log,
                    prior=hp.default_value if use_priors else None,
                    prior_confidence=std_to_confidences[std],
                )
            case CategoricalHyperparameter():
                assert hp.weights is None, (
                    "Weights on categoricals are not yet supported!"
                )
                space[hp.name] = neps.Categorical(
                    choices=hp.choices,
                    prior=hp.default_value if use_priors else None,
                    prior_confidence=std_to_confidences[std]
                )
            case OrdinalHyperparameter():
                space[hp.name] = neps.Categorical(
                    choices=hp.sequence,
                    prior=hp.default_value if use_priors else None,
                    prior_confidence=std_to_confidences[std]
                )
            case Constant():
                space[hp.name] = neps.Constant(
                    value=hp.value,
                )
    return neps.SearchSpace(space)



from __future__ import annotations

import argparse
import logging

from hposuite.benchmarks import BENCHMARKS, MF_BENCHMARKS, NON_MF_BENCHMARKS

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

def main(
    list: str
):
    match list:
        case "all":
            logger.info(BENCHMARKS.keys())
        case "mf":
            logger.info(MF_BENCHMARKS.keys())
        case "nonmf":
            logger.info(NON_MF_BENCHMARKS.keys())
        # TODO: Implement the following cases
        case "mo":
            raise NotImplementedError("No Multi-objective benchmark definitions exist yet")
        case "so":
            raise NotImplementedError("No Single-objective benchmark definitions exist yet")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--list",
        required=True,
        choices=["all", "mf", "nonmf", "mo", "so"],
        help="List all Benchmarks or filter by type"
        "\nall: list all Benchmarks"
        "\nmf: list multi-fidelity Benchmarks"
        "\nnonmf: list non-multi-fidelity Benchmarks"
        "\nmo: list multi-objective Benchmarks"
        "\nso: list single-objective Benchmarks",
    )
    args = parser.parse_args()
    main(args.list)



"""Script to interface with the MF Prior Bench library."""
# TODO(eddiebergman): Right now it's not clear how to set defaults for multi-objective.
# Do we want to prioritize obj and cost (i.e. accuracy and time) or would we rather two
# objectives (i.e. accuracy and cross entropy)?
# Second, for a benchmark to be useful, it should provide a reference point from which to compute
# hypervolume. For bounded costs this is fine but we can't do so for something like time.
# For tabular datasets, we could manually look for the worst time value
# TODO(eddiebergman): Have not included any of the conditional benchmarks for the moment
# as it seems to crash
# > "nb301": NB301Benchmark,
# > "rbv2_super": RBV2SuperBenchmark,
# > "rbv2_aknn": RBV2aknnBenchmark,
# > "rbv2_glmnet": RBV2glmnetBenchmark,
# > "rbv2_ranger": RBV2rangerBenchmark,
# > "rbv2_rpart": RBV2rpartBenchmark,
# > "rbv2_svm": RBV2svmBenchmark,
# > "rbv2_xgboost": RBV2xgboostBenchmark,
# > "iaml_glmnet": IAMLglmnetBenchmark,
# > "iaml_ranger": IAMLrangerBenchmark,
# > "iaml_rpart": IAMLrpartBenchmark,
# > "iaml_super": IAMLSuperBenchmark,
# > "iaml_xgboost": IAMLxgboostBenchmark,

from __future__ import annotations

import logging
import os
import warnings
from collections.abc import Iterator
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
from hpoglue import BenchmarkDescription, Measure, Result, SurrogateBenchmark
from hpoglue.env import Env
from hpoglue.fidelity import RangeFidelity

from hposuite.constants import DATA_DIR
from hposuite.utils import is_package_installed

mfp_logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    import mfpbench
    from hpoglue import Query


warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)


pd1_benchmarks = (
    "cifar100_wideresnet_2048",
    "imagenet_resnet_512",
    "lm1b_transformer_2048",
    "translatewmt_xformer_64",
)


def _get_surrogate_benchmark(
    description: BenchmarkDescription,
    *,
    benchmark_name: str,
    datadir: Path | None = None,
    **kwargs: Any,
) -> SurrogateBenchmark:
    import mfpbench

    if datadir is not None:
        datadir = datadir.resolve()
        if benchmark_name in pd1_benchmarks:
            if "pd1" in os.listdir(datadir):
                datadir = datadir / "pd1"
            else:
                raise ValueError(
                    f"Could not find pd1-{benchmark_name} Benchmark data in {datadir}. "
                    "Download the benchmark data using the command: \n"
                    f'python -m mfpbench download --benchmark "pd1" --data-dir {datadir}'
                )
        kwargs["datadir"] = datadir
    bench = mfpbench.get(benchmark_name, **kwargs)
    query_function = partial(_mfpbench_surrogate_query_function, benchmark=bench)
    return SurrogateBenchmark(
        desc=description,
        benchmark=bench,
        config_space=bench.space,
        query=query_function,
    )


def _mfpbench_surrogate_query_function(query: Query, benchmark: mfpbench.Benchmark) -> Result:
    if query.fidelity is not None:
        assert isinstance(query.fidelity, tuple)
        _, fid_value = query.fidelity
    else:
        fid_value = None
    return Result(
        query=query,
        values=benchmark.query(
            query.config.values,
            at=fid_value,
        ).as_dict(),
        fidelity=query.fidelity,
    )


def mfh() -> Iterator[BenchmarkDescription]:
    """Generates benchmark descriptions for the MF-Hartmann Benchmarks.

    Args:
        datadir (Path | None): The directory where the data is stored.
        If None, the default directory is used.

    Yields:
        Iterator[BenchmarkDescription]: An iterator over BenchmarkDescription objects
        for each combination of correlation and dimensions in the MFH Benchmarks.
    """
    import mfpbench
    env = Env(
        name="py310-mfpbench-1.10-mfh",
        python_version="3.10",
        requirements=("mf-prior-bench>=1.10.0",),
        post_install=(),
    )
    for req in env.requirements:
        if not is_package_installed(req):
            mfp_logger.warning(f"Please install the required package for mfh: {req}", stacklevel=2)
            return
    for correlation in ("bad", "good", "moderate", "terrible"):
        for dims in (3, 6):
            name = f"mfh{dims}_{correlation}"
            _min = -3.32237 if dims == 3 else -3.86278  # noqa: PLR2004
            yield BenchmarkDescription(
                name=name,
                config_space=mfpbench.get(name).space,
                load=partial(_get_surrogate_benchmark, benchmark_name=name),
                costs={
                    "fid_cost": Measure.cost((0.05, 1), minimize=True),
                },
                fidelities={
                    "z": RangeFidelity.from_tuple((1, 100, 1), supports_continuation=True),
                },
                metrics={
                    "value": Measure.metric((_min, np.inf), minimize=True),
                },
                has_conditionals=False,
                is_tabular=False,
                env=env,
                mem_req_mb = 1024,
            )


def pd1(datadir: Path) -> Iterator[BenchmarkDescription]:
    """Generates benchmark descriptions for the PD1 Benchmarks.

    Args:
        datadir (Path | None): The directory where the data is stored.
        If None, the default directory is used.

    Yields:
        Iterator[BenchmarkDescription]: An iterator over BenchmarkDescription objects
        for each PD1 benchmark.
    """
    import mfpbench
    env = Env(
        name="py310-mfpbench-1.10-pd1",
        python_version="3.10",
        requirements=("mf-prior-bench[pd1]>=1.10.0",),
        post_install=_download_data_cmd("pd1", datadir=datadir),
    )
    for req in env.requirements:
        if not is_package_installed(req):
            mfp_logger.warning(f"Please install the required package for pd1: {req}", stacklevel=2)
            return
    yield BenchmarkDescription(
        name="pd1-cifar100-wide_resnet-2048",
        config_space=mfpbench.pd1.benchmarks.PD1cifar100_wideresnet_2048._create_space(),
        load=partial(
            _get_surrogate_benchmark, benchmark_name="cifar100_wideresnet_2048", datadir=datadir
        ),
        metrics={"valid_error_rate": Measure.metric(bounds=(0, 1), minimize=True)},
        test_metrics={"test_error_rate": Measure.metric(bounds=(0, 1), minimize=True)},
        costs={"train_cost": Measure.cost(bounds=(0, np.inf), minimize=True)},
        fidelities={"epoch": RangeFidelity.from_tuple((1, 199, 1), supports_continuation=True)},
        is_tabular=False,
        has_conditionals=False,
        env=env,
        mem_req_mb=12288,
    )
    yield BenchmarkDescription(
        name="pd1-imagenet-resnet-512",
        config_space=mfpbench.pd1.benchmarks.PD1imagenet_resnet_512._create_space(),
        load=partial(
            _get_surrogate_benchmark, benchmark_name="imagenet_resnet_512", datadir=datadir
        ),
        metrics={"valid_error_rate": Measure.metric(bounds=(0, 1), minimize=True)},
        test_metrics={"test_error_rate": Measure.metric(bounds=(0, 1), minimize=True)},
        costs={"train_cost": Measure.cost(bounds=(0, np.inf), minimize=True)},
        fidelities={"epoch": RangeFidelity.from_tuple((1, 99, 1), supports_continuation=True)},
        is_tabular=False,
        has_conditionals=False,
        env=env,
        mem_req_mb=12288,
    )
    yield BenchmarkDescription(
        name="pd1-lm1b-transformer-2048",
        config_space=mfpbench.pd1.benchmarks.PD1lm1b_transformer_2048._create_space(),
        load=partial(
            _get_surrogate_benchmark, benchmark_name="lm1b_transformer_2048", datadir=datadir
        ),
        metrics={"valid_error_rate": Measure.metric(bounds=(0, 1), minimize=True)},
        test_metrics={"test_error_rate": Measure.metric(bounds=(0, 1), minimize=True)},
        costs={"train_cost": Measure.cost(bounds=(0, np.inf), minimize=True)},
        fidelities={"epoch": RangeFidelity.from_tuple((1, 74, 1), supports_continuation=True)},
        is_tabular=False,
        has_conditionals=False,
        env=env,
        mem_req_mb=24576,
    )
    yield BenchmarkDescription(
        name="pd1-translate_wmt-xformer_translate-64",
        config_space=mfpbench.pd1.benchmarks.PD1translatewmt_xformer_64._create_space(),
        load=partial(
            _get_surrogate_benchmark, benchmark_name="translatewmt_xformer_64", datadir=datadir
        ),
        metrics={"valid_error_rate": Measure.metric(bounds=(0, 1), minimize=True)},
        test_metrics={"test_error_rate": Measure.metric(bounds=(0, 1), minimize=True)},
        costs={"train_cost": Measure.cost(bounds=(0, np.inf), minimize=True)},
        fidelities={"epoch": RangeFidelity.from_tuple((1, 19, 1), supports_continuation=True)},
        is_tabular=False,
        has_conditionals=False,
        env=env,
        mem_req_mb=24576,
    )


def _download_data_cmd(key: str, datadir: Path | None = None) -> tuple[str, ...]:
    install_cmd = f"python -m mfpbench download --benchmark {key}"
    if datadir is not None:
        install_cmd += f" --data-dir {datadir.resolve()}"
    return tuple(install_cmd.split(" "))


def mfpbench_benchmarks(datadir: Path | None = None) -> Iterator[BenchmarkDescription]:
    """Generates benchmark descriptions for various MF-Prior-Bench.

    Args:
        datadir (Path | None): The directory where the data is stored.
        If None, the default directory is used.

    Yields:
        Iterator[BenchmarkDescription]: An iterator over BenchmarkDescription objects
        for each benchmark.
    """
    if isinstance(datadir, str):
        datadir = Path(datadir).resolve()
    elif datadir is None:
        datadir = DATA_DIR

    yield from mfh()
    yield from pd1(datadir)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from ConfigSpace import ConfigurationSpace
from hpoglue import Config, FunctionalBenchmark, Measure, Result

if TYPE_CHECKING:

    from hpoglue import Query


def ackley_fn(x: np.ndarray) -> float:
    """Compute the Ackley function.

    The Ackley function is a widely used benchmark function for testing optimization algorithms.
    It is defined as follows:

        f(x) = -a * exp(-b * sqrt(1/n * sum(x_i^2))) - exp(1/n * sum(cos(c * x_i))) + a + exp(1)

    where:
        - x is a numpy array of input values.
        - n is the number of dimensions (variables), which is set to 2 in this implementation.
        - a, b, and c are constants with typical values a=20, b=0.2, and c=2*pi.

    Parameters:
    x (np.ndarray): Input array of shape (n_var,).

    Returns:
    float: The computed value of the Ackley function.
    """
    n_var=2
    a=20
    b=1/5
    c=2 * np.pi
    part1 = -1. * a * np.exp(-1. * b * np.sqrt((1. / n_var) * np.sum(x * x)))
    part2 = -1. * np.exp((1. / n_var) * np.sum(np.cos(c * x)))

    return part1 + part2 + a + np.exp(1)


def wrapped_ackley(query: Query) -> Result:  # noqa: D103

    y = ackley_fn(
        np.array(query.config.to_tuple())
    )

    return Result(
        query=query,
        fidelity=None,
        values={"y": y},
    )


ACKLEY_BENCH = FunctionalBenchmark(
    name="ackley",
    config_space=ConfigurationSpace(
        {
            f"x{i}": (-32.768, 32.768) for i in range(2)
        }
    ),
    metrics={"y": Measure.metric((0.0, np.inf), minimize=True)},
    query=wrapped_ackley,
    predefined_points={
        "min": (
            Config(
                config_id="min",
                description="This point yields a global optimum of y:0.0",
                values={"x0": 0.0, "x1": 0.0}
            )
        )
    }
)


def branin_fn(x: np.ndarray) -> float:
    """Compute the value of the Branin function.

    The Branin function is a commonly used test function for optimization algorithms.
    It is defined as:

        f(x) = a * (x2 - b * x1^2 + c * x1 - r)^2 + s * (1 - t) * cos(x1) + s

    where:
        b = 5.1 / (4.0 * pi^2)
        c = 5.0 / pi
        t = 1.0 / (8.0 * pi)

    Args:
        x (np.ndarray): A 2-dimensional input array where x[0] is x1 and x[1] is x2.

    Returns:
        float: The computed value of the Branin function.
    """
    x1 = x[0]
    x2 = x[1]
    a = 1.0
    b = 5.1 / (4.0 * np.pi**2)
    c = 5.0 / np.pi
    r = 6.0
    s = 10.0
    t = 1.0 / (8.0 * np.pi)

    return a * (x2 - b * x1**2 + c * x1 - r) ** 2 + s * (1 - t) * np.cos(x1) + s


def wrapped_branin(query: Query) -> Result:  # noqa: D103

    y = branin_fn(
        np.array(query.config.to_tuple())
    )

    return Result(
        query=query,
        fidelity=None,
        values={"value": y},
    )


BRANIN_BENCH = FunctionalBenchmark(
    name="branin",
    config_space=ConfigurationSpace(
        {
            f"x{i}": (-32.768, 32.768) for i in range(2)
        }
    ),
    metrics={
            "value": Measure.metric((0.397887, np.inf), minimize=True),
        },
    query=wrapped_branin,
    predefined_points={
        "min": (
            Config(
                config_id="min",
                description="This point yields a global optimum of y:0.39787",
                values={"x0": -np.pi, "x1": 12.275},
            )
        ),
        "min2": (
            Config(
                config_id="min2",
                description="This point yields a global optimum of y:0.39787",
                values={"x0": np.pi, "x1": 2.275},
            )
        ),
        "min3": (
            Config(
                config_id="min3",
                description="This point yields a global optimum of y:0.39787",
                values={"x0": 9.42478, "x1": 2.475},
            )
        )
    }
)


"""hposuite interface for SMAC Optimizers."""

from __future__ import annotations

from collections.abc import Hashable, Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, NoReturn

from ConfigSpace import ConfigurationSpace
from hpoglue import Config, Optimizer, Problem, Query
from hpoglue.budget import CostBudget, TrialBudget
from hpoglue.env import Env
from smac import (
    AlgorithmConfigurationFacade,
    BlackBoxFacade,
    HyperbandFacade,
    HyperparameterOptimizationFacade,
    MultiFidelityFacade,
    Scenario,
)
from smac.acquisition.function import (
    EI,
    EIPS,
    LCB,
    PI,
    TS,
    PriorAcquisitionFunction,
)
from smac.runhistory import StatusType, TrialValue

from hposuite.utils import set_priors_as_defaults

if TYPE_CHECKING:
    from hpoglue import Result
    from hpoglue.fidelity import Fidelity
    from smac.facade import AbstractFacade
    from smac.runhistory import TrialInfo


acq_funcs = {
    "EI": EI,
    "EIPS": EIPS,
    "LCB": LCB,
    "PI": PI,
    "TS": TS,
}


bo_facades = {
    "AlgConf": AlgorithmConfigurationFacade,
    "BlackBox": BlackBoxFacade,
    "HPO": HyperparameterOptimizationFacade,
}

def _dummy_target_function(*args: Any, budget: int | float, seed: int) -> NoReturn:  # noqa: ARG001
    raise RuntimeError("This should never be called!")


class SMAC_Optimizer(Optimizer):
    """Default SMAC Optimizer."""

    env = Env(
        name="SMAC-2.1",
        python_version="3.10",
        requirements=("smac==2.1",)
    )


    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        optimizer: AbstractFacade,
        fidelity: Fidelity | None,
    ):
        """Create a SMAC Optimizer instance for a given problem.

        Args:
            problem: The problem to optimize.

            seed: Random seed for the optimizer.

            working_directory: Working directory to store SMAC run.

            config_space: Configuration space to optimize over.

            optimizer: SMAC optimizer instance.

            fidelity: Fidelity space to optimize over, if any.
        """
        self.problem = problem
        self.working_directory = working_directory
        self.optimizer = optimizer
        self.config_space = problem.config_space
        self._trial_lookup: dict[Hashable, TrialInfo] = {}
        self._fidelity = fidelity
        self._seed = seed

    def ask(self) -> Query:
        """Ask SMAC for a new config to evaluate."""
        smac_info = self.optimizer.ask()
        assert smac_info.instance is None, "We don't do instance benchmarks!"

        config = smac_info.config
        raw_config = dict(config)
        config_id = str(self.optimizer.intensifier.runhistory.config_ids[config])

        fidelity = smac_info.budget

        match fidelity:
            case None:
                fidelity = None
            case float() | int():
                assert self._fidelity is not None
                assert isinstance(self.problem.fidelities, tuple)
                fidelity_value = int(fidelity) if self._fidelity.kind is int else fidelity
                fidelity_name = self.problem.fidelities[0]
                fidelity = (fidelity_name, fidelity_value)
            case _:
                raise NotImplementedError("Unexpected return type for SMAC budget!")

        return Query(
            config=Config(config_id=config_id, values=raw_config),
            fidelity=fidelity,
            optimizer_info=smac_info,
        )

    def tell(self, result: Result) -> None:
        """Tell SMAC the result of the query."""
        match self.problem.objectives:
            case Mapping():
                cost = [
                    obj.as_minimize(result.values[key])
                    for key, obj in self.problem.objectives.items()
                ]
            case (key, obj):
                cost = obj.as_minimize(result.values[key])
            case _:
                raise TypeError("Objective must be a string or a list of strings")

        self.optimizer.tell(
            result.query.optimizer_info,  # type: ignore
            TrialValue(
                cost=cost,
                time=0.0,
                starttime=0.0,
                endtime=0.0,
                status=StatusType.SUCCESS,
                additional_info={},
            ),
            save=True,
        )


class SMAC_BO(SMAC_Optimizer):
    """SMAC Bayesian Optimization."""

    name = "SMAC_BO"

    support = Problem.Support(
        fidelities=(None,),
        objectives=("single", "many"),
        cost_awareness=(None,),
        tabular=False,
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        facade: Literal[
            "HPO",
            "BlackBox",
            "AlgConf",
        ] = "BlackBox",
        acq_func: Literal[
            "EI",
            "EIPS",
            "LCB",
            "PI",
            "TS",
        ] = "EI",
        acq_func_kwargs: dict[str, Any] | None = None,
    ):
        """Create a SMAC BO Optimizer instance for a given problem.

        Args:
            problem: The problem to optimize.

            seed: Random seed for the optimizer.

            working_directory: Working directory to store SMAC run.

            facade: The SMAC facade to use.
                - "HPO": HyperparameterOptimizationFacade
                - "BlackBox": BlackBoxFacade
                - "AlgConf": AlgorithmConfigurationFacade
            Defaults to "BlackBox".

            acq_func: The acquisition function to use.
                - "EI": Expected Improvement
                - "EIPS": Expected Improvement per Second
                - "LCB": Lower Confidence Bound
                - "PI": Probability of Improvement
                - "TS": Thompson Sampling

            acq_func_kwargs: Additional arguments for the acquisition function.
                Defaults to None.
                See SMAC documentation for details.
        """
        assert facade in bo_facades, f"Unknown facade for SMAC_BO: {facade}"
        assert acq_func in acq_funcs, f"Unknown acquisition function for SMAC_BO: {acq_func}"

        config_space = problem.config_space
        match config_space:
            case ConfigurationSpace():
                pass
            case list():
                raise ValueError("SMAC does not support tabular benchmarks!")
            case _:
                raise TypeError("Config space must be a list or a ConfigurationSpace!")

        match problem.fidelities:
            case None:
                pass
            case tuple() | Mapping():
                raise ValueError("SMAC BO does not support multi-fidelity benchmarks!")
            case _:
                raise TypeError("Fidelity must be a string or a list of strings!")

        match problem.objectives:
            case Mapping():
                metric_names = list(problem.objectives.keys())
            case (metric_name, _):
                metric_names = metric_name
            case _:
                raise TypeError("Objective must be a tuple of (name, metric) or a mapping")

        working_directory.mkdir(parents=True, exist_ok=True)

        match problem.budget:
            case TrialBudget():
                budget = problem.budget.total
            case CostBudget():
                raise ValueError("SMAC BO does not support cost-aware benchmarks!")
            case _:
                raise TypeError("Budget must be a TrialBudget or a CostBudget!")

        scenario = Scenario(
            configspace=config_space,
            deterministic=True,
            objectives=metric_names,
            n_trials=budget,
            seed=seed,
            output_directory=working_directory / "smac-output",
            min_budget=None,
            max_budget=None,
        )

        facade: AbstractFacade = bo_facades[facade]

        acquisition_func = acq_funcs[acq_func](
            **acq_func_kwargs
        )

        super().__init__(
            problem=problem,
            seed=seed,
            working_directory=working_directory,
            fidelity=None,
            optimizer=facade(
                scenario=scenario,
                logging_level=False,
                target_function=_dummy_target_function,
                intensifier=facade.get_intensifier(scenario),
                acquisition_function=acquisition_func,
                overwrite=True,
            ),
        )


class SMAC_Hyperband(SMAC_Optimizer):
    """SMAC Hyperband."""

    name = "SMAC_Hyperband"
    support = Problem.Support(
        fidelities=("single",),
        objectives=("single", "many"),
        cost_awareness=(None,),
        tabular=False,
        continuations=True,
    )
    mem_req_mb = 1024

    def __init__(  # noqa: PLR0912
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        eta: int = 3,
    ):
        """Create a SMAC Hyperband Optimizer instance for a given problem.

        Args:
            problem: The problem to optimize.

            seed: Random seed for the optimizer.

            working_directory: Working directory to store SMAC run.

            eta: Sets the aggressiveness of Hyperband's aggressive
                early stopping by retaining 1/eta configurations every round.
                Defaults to 3.
        """
        config_space = problem.config_space
        match config_space:
            case ConfigurationSpace():
                pass
            case list():
                raise ValueError("SMAC does not support tabular benchmarks!")
            case _:
                raise TypeError("Config space must be a list or a ConfigurationSpace!")

        min_fidelity: float | int
        max_fidelity: float | int
        match problem.fidelities:
            case None:
                raise ValueError("SMAC Hyperband requires a fidelity space!")
            case Mapping():
                raise ValueError("SMAC Hyperband does not support many-fidelity!")
            case (_, fidelity):
                _fid = fidelity
                min_fidelity = fidelity.min
                max_fidelity = fidelity.max
            case _:
                raise TypeError("Fidelity must be a string or a list of strings!")

        match problem.objectives:
            case Mapping():
                metric_names = list(problem.objectives.keys())
            case (metric_name, _):
                metric_names = metric_name
            case _:
                raise TypeError("Objective must be a tuple of (name, metric) or a mapping")

        working_directory.mkdir(parents=True, exist_ok=True)
        match problem.budget:
            case TrialBudget():
                budget = problem.budget.total
            case CostBudget():
                raise ValueError("SMAC Hyperband does not support cost-aware benchmarks!")
            case _:
                raise TypeError("Budget must be a TrialBudget or a CostBudget!")

        scenario = Scenario(
            configspace=config_space,
            deterministic=True,
            objectives=metric_names,
            n_trials=budget,
            seed=seed,
            output_directory=working_directory / "smac-output",
            min_budget=min_fidelity,
            max_budget=max_fidelity,
        )
        super().__init__(
            problem=problem,
            seed=seed,
            working_directory=working_directory,
            fidelity=_fid,
            optimizer=HyperbandFacade(
                scenario=scenario,
                logging_level=False,
                target_function=_dummy_target_function,
                intensifier=HyperbandFacade.get_intensifier(scenario, eta=eta),
                overwrite=True,
            ),
        )


class SMAC_BOHB(SMAC_Optimizer):
    """SMAC BOHB."""

    name = "SMAC_BOHB"
    support = Problem.Support(
        fidelities=("single",),
        objectives=("single", "many"),
        cost_awareness=(None,),
        tabular=False,
        continuations=True,
    )
    mem_req_mb = 1024

    def __init__(  # noqa: PLR0912
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        eta: int = 3,
    ):
        """Create a SMAC BOHB Optimizer instance for a given problem.

        Args:
            problem: The problem to optimize.

            seed: Random seed for the optimizer.

            working_directory: Working directory to store SMAC run.

            eta: Sets the aggressiveness of Hyperband's aggressive
                early stopping by retaining 1/eta configurations every round.
                Defaults to 3.
                According to the paper, changing this doesn't seem to make a difference.
        """
        config_space = problem.config_space
        match config_space:
            case ConfigurationSpace():
                pass
            case list():
                raise ValueError("SMAC does not support tabular benchmarks!")
            case _:
                raise TypeError("Config space must be a list or a ConfigurationSpace!")

        min_fidelity: float | int
        max_fidelity: float | int
        match problem.fidelities:
            case None:
                raise ValueError("SMAC BOHB requires a fidelity space!")
            case Mapping():
                raise ValueError("SMAC BOHB does not support many-fidelity!")
            case (_, fidelity):
                _fid = fidelity
                min_fidelity = fidelity.min
                max_fidelity = fidelity.max
            case _:
                raise TypeError("Fidelity must be a string or a list of strings!")

        match problem.objectives:
            case Mapping():
                metric_names = list(problem.objectives.keys())
            case (metric_name, _):
                metric_names = metric_name
            case _:
                raise TypeError("Objective must be a tuple of (name, metric) or a mapping")

        working_directory.mkdir(parents=True, exist_ok=True)
        match problem.budget:
            case TrialBudget():
                budget = problem.budget.total
            case CostBudget():
                raise ValueError("SMAC BOHB does not support cost-aware benchmarks!")
            case _:
                raise TypeError("Budget must be a TrialBudget or a CostBudget!")

        scenario = Scenario(
            configspace=config_space,
            deterministic=True,
            objectives=metric_names,
            n_trials=budget,
            seed=seed,
            output_directory=working_directory / "smac-output",
            min_budget=min_fidelity,
            max_budget=max_fidelity,
        )
        super().__init__(
            problem=problem,
            seed=seed,
            working_directory=working_directory,
            fidelity=_fid,
            optimizer=MultiFidelityFacade(
                scenario=scenario,
                logging_level=False,
                target_function=_dummy_target_function,
                intensifier=MultiFidelityFacade.get_intensifier(scenario, eta=eta),
                overwrite=True,
            ),
        )


class SMAC_PiBO(SMAC_Optimizer):
    """πBO Optimizer in SMAC."""

    name = "SMAC_PiBO"

    support = Problem.Support(
        fidelities=(None,),
        objectives=("single"),
        cost_awareness=(None,),
        tabular=False,
        priors=True,
    )

    mem_req_mb = 1024

    def __init__(  # noqa: C901, PLR0912
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        decay_beta: float = 10.0,
        facade: Literal[
            "HPO",
            "BlackBox",
            "AlgConf",
        ] = "HPO",
        acq_func: Literal[
            "EI",
            "EIPS",
            "LCB",
            "PI",
            "TS",
        ] = "EI",
        acq_func_kwargs: dict[str, Any] | None = None,
    ):
        """Create a SMAC BO Optimizer instance for a given problem.

        Args:
            problem: The problem to optimize.

            seed: Random seed for the optimizer.

            working_directory: Working directory to store SMAC run.

            decay_beta: Parameter to control the decay of the prior strength
            in the acquisition function.

            facade: The SMAC facade to use.
                - "HPO": HyperparameterOptimizationFacade
                - "BlackBox": BlackBoxFacade
                - "AlgConf": AlgorithmConfigurationFacade
            Defaults to "HPO".

            acq_func: The acquisition function to use.
                - "EI": Expected Improvement
                - "EIPS": Expected Improvement per Second
                - "LCB": Lower Confidence Bound
                - "PI": Probability of Improvement
                - "TS": Thompson Sampling
            acq_func_kwargs: Additional arguments for the acquisition function.
                Defaults to None.
                See SMAC documentation for details.
        """
        assert facade in bo_facades, f"Unknown facade for SMAC_PiBO: {facade}"
        assert acq_func in acq_funcs, f"Unknown acquisition function for SMAC_PiBO: {acq_func}"

        config_space = problem.config_space
        match config_space:
            case ConfigurationSpace():
                pass
            case list():
                raise ValueError("SMAC does not support tabular benchmarks!")
            case _:
                raise TypeError("Config space must be a list or a ConfigurationSpace!")

        match problem.fidelities:
            case None:
                pass
            case tuple() | Mapping():
                raise ValueError("SMAC BO does not support multi-fidelity benchmarks!")
            case _:
                raise TypeError("Fidelity must be a string or a list of strings!")

        match problem.objectives:
            case Mapping():
                metric_names = list(problem.objectives.keys())
            case (metric_name, _):
                metric_names = metric_name
            case _:
                raise TypeError("Objective must be a tuple of (name, metric) or a mapping")

        working_directory.mkdir(parents=True, exist_ok=True)

        match problem.budget:
            case TrialBudget():
                budget = problem.budget.total
            case CostBudget():
                raise ValueError("SMAC BO does not support cost-aware benchmarks!")
            case _:
                raise TypeError("Budget must be a TrialBudget or a CostBudget!")

        match problem.priors:
            case None:
                raise ValueError("SMAC PiBO requires priors!")
            case tuple():
                assert len(problem.priors[1]) == 1, (
                    "SMAC PiBO doesn't support multi-objective priors! "
                )
                prior = next(iter(problem.priors[1].values()))
            case _:
                raise TypeError(
                    "Priors must be a tuple of (name, dict(objective, Config). "
                    f"Got {type(problem.priors)}"
                )

        config_space = set_priors_as_defaults(
            config_space=config_space,
            priors=prior,
        )

        scenario = Scenario(
            configspace=config_space,
            deterministic=True,
            objectives=metric_names,
            n_trials=budget,
            seed=seed,
            output_directory=working_directory / "smac-output",
            min_budget=None,
            max_budget=None,
        )

        facade: AbstractFacade = bo_facades[facade]

        acquisition_func = acq_funcs[acq_func](
            **acq_func_kwargs
        )

        super().__init__(
            problem=problem,
            seed=seed,
            working_directory=working_directory,
            fidelity=None,
            optimizer=facade(
                scenario=scenario,
                logging_level=False,
                target_function=_dummy_target_function,
                intensifier=facade.get_intensifier(scenario),
                acquisition_function=PriorAcquisitionFunction(
                    acquisition_function=acquisition_func,
                    decay_beta=scenario.n_trials / decay_beta,
                ),
                overwrite=True,
            ),
        )


"""hposuite interface for Scikit_Optimize."""

from __future__ import annotations

import logging
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING

import ConfigSpace as CS  # noqa: N817
import skopt
from hpoglue import Config, Optimizer, Problem, Query
from hpoglue.env import Env

if TYPE_CHECKING:
    from hpoglue import Result
    from skopt.space.space import Space

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


base_estimators = ["GP", "RF", "ET", "GBRT"]
acq_funcs = ["LCB", "EI", "PI", "gp_hedge"]
acq_optimizers = ["sampling", "lbfgs", "auto"]

class SkoptOptimizer(Optimizer):
    """The Scikit_Optimize Optimizer."""

    name = "Scikit_Optimize"

    env = Env(
        name="scikit_optimize-0.10.2",
        python_version="3.10",
        requirements=("scikit-optimize==0.10.2",),
    )

    support = Problem.Support(
        fidelities=(None,),  # NOTE: Skopt does not support multi-fidelity optimization
        objectives=("single", "many"),
        cost_awareness=(None,),
        tabular=False,
    )

    mem_req_mb = 1024


    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        base_estimator: str = "GP",
        acq_func: str = "gp_hedge",
        acq_optimizer: str = "auto",
    ) -> None:
        """Create an Skopt Optimizer instance for a given problem."""
        self.config_space = problem.config_space
        self._space: list[Space]
        match self.config_space:
            case CS.ConfigurationSpace():
                self._space = _configspace_to_skopt_space(self.config_space)
            case list():
                raise ValueError("SciKit-Optimize does not support list-type config spaces!")
            case _:
                raise TypeError("Config space must be a list or a ConfigurationSpace!")

        assert base_estimator in base_estimators, f"base_estimator must be one of {base_estimators}"

        assert acq_func in acq_funcs, f"acq_func must be one of {acq_funcs}"

        assert acq_optimizer in acq_optimizers, f"acq_optimizer must be one of {acq_optimizers}"

        self.optimizer: skopt.optimizer.Optimizer
        self.optimizer = skopt.optimizer.Optimizer(
            dimensions=self._space,
            base_estimator=base_estimator,
            acq_func=acq_func,
            acq_optimizer=acq_optimizer,
            random_state=seed,
            n_initial_points=5
        )

        self.problem = problem
        self.working_directory = working_directory
        self.trial_counter = 0


    def ask(self) -> Query:
        """Get a new config from the optimizer."""
        match self.problem.fidelities:
            case None:
                config = self.optimizer.ask()
                config_values = {
                    hp.name: value
                    for hp, value in zip(
                        self.config_space.get_hyperparameters(),
                        config,
                        strict=False
                    )
                }
                assert list(config_values.keys()) == \
                    list(self.config_space.get_hyperparameter_names())
                assert list(config_values.keys()) == [hp.name for hp in self._space]
                name = f"trial_{self.trial_counter}"
                self.trial_counter += 1
                return Query(
                    config=Config(config_id=name, values=config_values),
                    fidelity=None,
                    optimizer_info=None,
                )
            case tuple():
                raise ValueError("Multi-fidelity optimization not supported by Scikit_Optimize!")
            case Mapping():
                raise ValueError("Many-fidelity optimization not supported by Scikit_Optimize!")
            case _:
                raise TypeError("Fidelity must be None, a tuple or a Mapping!")


    def tell(self, result: Result) -> None:
        """Tell the optimizer about the result of a query."""
        match self.problem.objectives:
            case (name, metric):
                _values = metric.as_minimize(result.values[name])
            case Mapping():
                raise ValueError("Multiobjective not supported by Scikit_Optimize!")
            case _:
                raise TypeError("Objective must be a string or a list of strings!")

        match self.problem.costs:
            case None:
                pass
            case tuple():
                raise ValueError("Cost-aware optimization not supported by Scikit_Optimize!")
            case Mapping():
                raise ValueError("Cost-aware optimization not supported by Scikit_Optimize!")
            case _:
                raise TypeError("Cost must be None or a mapping!")

        _ = self.optimizer.tell(
            list(result.query.config.values.values()),
            _values
        )


def _configspace_to_skopt_space(  # noqa: C901
    config_space: CS.ConfigurationSpace,
) -> dict[str, Space]:
    import numpy as np
    from skopt.space.space import Categorical, Integer, Real

    if len(config_space.conditions) > 0:
        raise NotImplementedError("Conditions are not yet supported!")

    if len(config_space.forbidden_clauses) > 0:
        raise NotImplementedError("Forbiddens are not yet supported!")

    skopt_space: list[float] = []
    for hp in list(config_space.values()):
        match hp:
            case CS.UniformIntegerHyperparameter() if hp.log:
                skopt_space.append(
                    Integer(
                        hp.lower,
                        hp.upper,
                        name=hp.name,
                        prior="log-uniform",
                        base=np.e,
                    )
                )
            case CS.UniformIntegerHyperparameter():
                skopt_space.append(Integer(hp.lower, hp.upper, name=hp.name))
            case CS.UniformFloatHyperparameter() if hp.log:
                skopt_space.append(
                    Real(
                        hp.lower,
                        hp.upper,
                        name=hp.name,
                        prior="log-uniform",
                        base=np.e,
                    )
                )
            case CS.UniformFloatHyperparameter():
                skopt_space.append(Real(hp.lower, hp.upper, name=hp.name))
            case CS.CategoricalHyperparameter() if hp.weights is not None:
                weights = np.asarray(hp.weights) / np.sum(hp.weights)
                skopt_space.append(Categorical(hp.choices, name=hp.name, prior=weights))
            case CS.CategoricalHyperparameter():
                skopt_space.append(Categorical(hp.choices, name=hp.name))
            case CS.Constant():
                skopt_space.append(Categorical([hp.value], name=hp.name))
            case CS.OrdinalHyperparameter():
                skopt_space.append(Categorical(list(hp.sequence), name=hp.name))
            case _:
                raise ValueError(
                    f"Unrecognized type of hyperparameter in ConfigSpace: {hp.__class__.__name__}!"
                )

    return skopt_space




"""hposuite interface for the Optuna Optimizers package."""

from __future__ import annotations

import logging
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING
from typing_extensions import Any

import ConfigSpace as CS  # noqa: N817
import optuna
from hpoglue import Config, Optimizer, Problem, Query
from hpoglue.env import Env

if TYPE_CHECKING:
    from hpoglue import Result
    from optuna.distributions import BaseDistribution

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class OptunaOptimizer(Optimizer):
    """The Optuna Optimizer."""

    name = "Optuna"

    env = Env(
        name="optuna-4.0.0",
        python_version="3.10",
        requirements=("optuna==4.0.0",),
    )


    support = Problem.Support(
        fidelities=(None,),  # TODO: Implement fidelity support
        objectives=("single", "many"),
        cost_awareness=(None,),
        tabular=False,
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        **kwargs: Any,
    ) -> None:
        """Create an Optuna Optimizer instance for a given problem."""
        import optuna
        from optuna.samplers import NSGAIISampler, TPESampler

        self._distributions: dict[str, BaseDistribution]
        config_space = problem.config_space
        match config_space:
            case CS.ConfigurationSpace():
                self._distributions = _configspace_to_optuna_distributions(config_space)
            case list():
                raise NotImplementedError("# TODO: Tabular not yet implemented for Optuna!")
            case _:
                raise TypeError("Config space must be a list or a ConfigurationSpace!")

        self.optimizer: optuna.study.Study
        match problem.objectives:
            case (_, objective):
                self.optimizer = optuna.create_study(
                    sampler=TPESampler(seed=seed, **kwargs),
                    storage=None,
                    pruner=None,  # TODO(eddiebergman): Figure out how to use this for MF
                    study_name=f"{problem.name}-{seed}",
                    load_if_exists=False,
                    direction="minimize" if objective.minimize else "maximize",
                )
            case Mapping():
                self.optimizer = optuna.create_study(
                    sampler=NSGAIISampler(seed=seed, **kwargs),
                    storage=None,
                    pruner=None,  # TODO(eddiebergman): Figure out how to use this for MF
                    study_name=f"{problem.name}-{seed}",
                    load_if_exists=False,
                    directions=[
                        "minimize" if obj.minimize else "maximize"
                        for obj in problem.objectives.values()
                    ],
                )
            case _:
                raise ValueError("Objective must be a string or a list of strings!")

        self.problem = problem
        self.working_directory = working_directory
        self._trial_lookup: dict[str, optuna.trial.Trial] = {}


    def ask(self) -> Query:
        """Ask the optimizer for a new configuration to evaluate."""
        match self.problem.fidelities:
            case None:
                trial = self.optimizer.ask(self._distributions)
                name = f"trial_{trial.number}"
                return Query(
                    config=Config(config_id=name, values=trial.params),
                    fidelity=None,
                    optimizer_info=trial,
                )
            case tuple():
                # TODO(eddiebergman): Not sure if just using
                # trial.number is enough in MF setting
                raise NotImplementedError("# TODO: Fidelity-aware not yet implemented for Optuna!")
            case Mapping():
                raise NotImplementedError("# TODO: Fidelity-aware not yet implemented for Optuna!")
            case _:
                raise TypeError("Fidelity must be None or a tuple!")


    def tell(self, result: Result) -> None:
        """Tell the optimizer about the result of a query."""
        match self.problem.objectives:
            case (name, metric):
                _values = metric.as_minimize(result.values[name])
            case Mapping():
                _values = [
                    metric.as_minimize(result.values[name])
                    for name, metric in self.problem.objectives.items()
                ]
            case _:
                raise TypeError("Objective must be a string or a list of strings!")

        match self.problem.costs:
            case None:
                pass
            case tuple():
                raise NotImplementedError("# TODO: Cost-aware not yet implemented for Optuna!")
            case Mapping():
                raise NotImplementedError("# TODO: Cost-aware not yet implemented for Optuna!")
            case _:
                raise TypeError("Cost must be None or a mapping!")

        assert isinstance(result.query.optimizer_info, optuna.trial.Trial)
        self.optimizer.tell(
            trial=result.query.optimizer_info,
            values=_values,
            state=optuna.trial.TrialState.COMPLETE,
            skip_if_finished=False,
        )


def _configspace_to_optuna_distributions(  # noqa: C901, PLR0912
    config_space: CS.ConfigurationSpace,
) -> dict[str, BaseDistribution]:
    from optuna.distributions import (
        CategoricalDistribution as Cat,
        FloatDistribution as Float,
        IntDistribution as Int,
    )

    from hposuite.utils import compare_installed_CS_version_vs_required

    if compare_installed_CS_version_vs_required(required_version="1.0") == "<":
        if len(config_space.get_conditions()) > 0:
            raise NotImplementedError("Conditions are not yet supported!")
        if len(config_space.get_forbiddens()) > 0:
            raise NotImplementedError("Forbiddens are not yet supported!")


    else:
        if len(config_space.conditions) > 0:
            raise NotImplementedError("Conditions are not yet supported!")

        if len(config_space.forbidden_clauses) > 0:
            raise NotImplementedError("Forbiddens are not yet supported!")

    optuna_space: dict[str, BaseDistribution] = {}
    for hp in list(config_space.values()):
        match hp:
            case CS.UniformIntegerHyperparameter():
                optuna_space[hp.name] = Int(hp.lower, hp.upper, log=hp.log)
            case CS.UniformFloatHyperparameter():
                optuna_space[hp.name] = Float(hp.lower, hp.upper, log=hp.log)
            case CS.CategoricalHyperparameter():
                if hp.weights is not None:
                    raise NotImplementedError("Weights on categoricals are not yet supported!")
                optuna_space[hp.name] = Cat(hp.choices)
            case CS.Constant():
                optuna_space[hp.name] = Cat([hp.value])
            case CS.OrdinalHyperparameter():
                raise NotImplementedError("Ordinal hyperparameters are not yet supported!")
            case _:
                raise ValueError("Unrecognized type of hyperparameter in ConfigSpace!")

    return optuna_space



from __future__ import annotations

import importlib
import logging
import types
from collections.abc import Generator, Iterator

from hpoglue import BenchmarkDescription, FunctionalBenchmark

from hposuite.exceptions import OptBenchNotInstalledError

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

modules = [
    ("hposuite.benchmarks.synthetic", "ACKLEY_BENCH", "BRANIN_BENCH"),
    ("hposuite.benchmarks.mfp_bench", "mfpbench_benchmarks"),
    ("hposuite.benchmarks.lcbench_tabular", "lcbench_tabular_benchmarks"),
    ("hposuite.benchmarks.pymoo", "pymoo_benchmarks"),
    ("hposuite.benchmarks.bbob_tabular", "bbob_tabular_benchmarks"),
    ("hposuite.benchmarks.mfh_tabular", "mfh_tabular_benchmarks"),
    ("hposuite.benchmarks.bbob", "bbob_benchmarks"),
]

imported_benches = []

for module_name, *attrs in modules:
    try:
        module = importlib.import_module(module_name)
        for attr in attrs:
            bench = getattr(module, attr)
            if isinstance(bench, types.FunctionType):
                bench = bench()
            match bench:
                case FunctionalBenchmark() | BenchmarkDescription():
                    imported_benches.append(bench)
                case Iterator() | Generator():
                    for b in bench:
                        match b:
                            case FunctionalBenchmark() | BenchmarkDescription():
                                imported_benches.append(b)
                            case _:
                                raise ValueError(f"Unexpected benchmark type: {type(b)}")
                case _:
                    raise ValueError(f"Unexpected benchmark type: {type(bench)}")
    except ImportError as e:
        logger.warning(OptBenchNotInstalledError(module_name, e.msg))


BENCHMARKS: dict[str, BenchmarkDescription] = {}
MF_BENCHMARKS: dict[str, BenchmarkDescription] = {}
NON_MF_BENCHMARKS: dict[str, BenchmarkDescription] = {}


for bench in imported_benches:
    BENCHMARKS[bench.name] = bench
    _bench = bench
    if isinstance(_bench, FunctionalBenchmark):
        _bench = _bench.desc

    if _bench.fidelities is not None:
        MF_BENCHMARKS[_bench.name] = _bench
    else:
        NON_MF_BENCHMARKS[_bench.name] = _bench

__all__ = [
    "BENCHMARKS",
    "MF_BENCHMARKS",
    "NON_MF_BENCHMARKS",
]
